#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2020 by the Linux Foundation
#
__author__ = 'Konstantin Ryabitsev <konstantin@linuxfoundation.org>'

import os
import sys
import mailbox
import email
import email.utils
import email.parser
import re
import time
import json
import fnmatch
import shutil
import pathlib
import io
import shlex
import argparse

import b4

from typing import Any, Optional, Union, List, Set, Dict, Tuple
from string import Template

from email.message import EmailMessage

logger = b4.logger

DEFAULT_MERGE_TEMPLATE = """Merge ${patch_or_series} "${seriestitle}"

${authorname} <${authoremail}> says:

${covermessage}

Link: ${midurl}
"""


def save_msgs_as_mbox(dest: str, msgs: List[EmailMessage], filterdupes: bool = False) -> int:
    if dest == '-':
        b4.save_mboxrd_mbox(msgs, sys.stdout.buffer, mangle_from=False)
        return len(msgs)

    if b4.is_maildir(dest):
        mdr = mailbox.Maildir(dest)
        have_msgids = set()
        added = 0
        if filterdupes:
            for emsg in mdr:
                have_msgids.add(b4.LoreMessage.get_clean_msgid(emsg))  # type: ignore[arg-type]
        for msg in msgs:
            if b4.LoreMessage.get_clean_msgid(msg) not in have_msgids:
                added += 1
                mdr.add(msg)
        return added

    with open(dest, 'wb') as fh:
        b4.save_mboxrd_mbox(msgs, fh, mangle_from=True)

    return len(msgs)


def get_base_commit(topdir: Optional[str], body: str, lser: b4.LoreSeries,
                    cmdargs: argparse.Namespace) -> str:
    base_commit = 'HEAD'

    if lser.prereq_base_commit:
        logger.debug('Setting base-commit to prereq-base-commit: %s', lser.prereq_base_commit)
        base_commit = lser.prereq_base_commit
    else:
        matches = re.search(r'base-commit: .*?([\da-f]+)', body, re.MULTILINE)
        if matches:
            base_commit = matches.groups()[0]
        else:
            # Try a more relaxed search
            matches = re.search(r'based on .*?([\da-f]{40})', body, re.MULTILINE)
            if matches:
                base_commit = matches.groups()[0]

    if base_commit and topdir:
        # Does it actually exist in this tree?
        if not b4.git_commit_exists(topdir, base_commit):
            logger.warning(' Base: base-commit %s not known, ignoring', base_commit)
            base_commit = 'HEAD'
        elif not cmdargs.mergebase:
            logger.debug(' Base: using specified base-commit %s', base_commit)

    if base_commit == 'HEAD' and topdir and cmdargs.guessbase:
        logger.info(' Base: attempting to guess base-commit...')
        try:
            base_commit, nblobs, mismatches = lser.find_base(topdir, branches=cmdargs.guessbranch,
                                                             maxdays=cmdargs.guessdays)
            if mismatches == 0:
                logger.critical(' Base: %s (exact match)', base_commit)
            elif nblobs == mismatches:
                logger.critical(' Base: failed to guess base')
            else:
                logger.critical(' Base: %s (best guess, %s/%s blobs matched)', base_commit,
                                nblobs - mismatches, nblobs)
        except IndexError as ex:
            logger.critical(' Base: failed to guess base (%s)', ex)

    if cmdargs.mergebase:
        if base_commit:
            logger.debug(' Base: overriding submitter provided base-commit %s with %s',
                           base_commit, cmdargs.mergebase)
        base_commit = cmdargs.mergebase

    return base_commit


def make_am(msgs: List[EmailMessage], cmdargs: argparse.Namespace, msgid: str) -> None:
    config = b4.get_main_config()
    outdir = cmdargs.outdir
    if outdir == '-':
        cmdargs.nocover = True
    if 'addmsgid' in cmdargs and cmdargs.addmsgid:
        logger.debug('Setting linktrailermask to Message-ID:')
        config['linktrailermask'] = 'Message-ID: <%s>'
        cmdargs.addlink = True
    wantver = cmdargs.wantver
    count = len(msgs)
    logger.info('Analyzing %s messages in the thread', count)
    lmbx = b4.LoreMailbox()
    # Go through the mbox once to populate base series
    load_codereview = True
    for msg in msgs:
        # Is it a collection of patches attached to the same message?
        # We only trigger this mode with --single-msg
        if msg.is_multipart() and ('singlemsg' in cmdargs and cmdargs.singlemsg):
            xpatches: List[EmailMessage] = list()
            xmsgid = b4.LoreMessage.get_clean_msgid(msg)
            for part in msg.walk():
                cte = part.get_content_type()
                if cte.find('/x-patch') < 0:
                    continue
                bpayload = part.get_payload(decode=True)
                if not isinstance(bpayload, bytes):
                    continue
                pcharset = part.get_content_charset()
                if not pcharset:
                    pcharset = 'utf-8'
                try:
                    payload = bpayload.decode(pcharset, errors='replace')
                except LookupError:
                    # what kind of encoding is that?
                    # Whatever, we'll use utf-8 and hope for the best
                    payload = bpayload.decode('utf-8', errors='replace')
                    part.set_param('charset', 'utf-8')
                if payload and b4.DIFF_RE.search(payload):
                    xmsg = email.parser.Parser(policy=b4.emlpolicy, _class=EmailMessage).parsestr(payload)
                    # Needs to have Subject, From, Date for us to consider it
                    if xmsg.get('Subject') and xmsg.get('From') and xmsg.get('Date'):
                        logger.debug('Found attached patch: %s', xmsg.get('Subject'))
                        xmsg['Message-ID'] = f'<att{len(xpatches)}-{xmsgid}>'
                        xpatches.append(xmsg)
            if len(xpatches):
                logger.info('Warning: Found %s patches attached to the requested message', len(xpatches))
                logger.info('         This mode ignores any follow-up trailers, use with caution')
                # Throw out lmbx and only use these
                lmbx = b4.LoreMailbox()
                load_codereview = False
                for xmsg in xpatches:
                    lmbx.add_message(xmsg)
                # Make a cover letter out of the original message
                cmsg = EmailMessage()
                cbody, ccharset = b4.LoreMessage.get_payload(msg, use_patch=False)
                cmsg['From'] = msg.get('From')
                cmsg['Date'] = msg.get('Date')
                cmsg['Message-ID'] = msg.get('Message-ID')
                cmsg['Subject'] = '[PATCH 0/0] ' + msg.get('Subject', '(no subject)')
                cmsg.set_payload(cbody, ccharset)
                lmbx.add_message(cmsg)
                break
        else:
            lmbx.add_message(msg)

    reroll = True
    if cmdargs.nopartialreroll:
        reroll = False

    lser = lmbx.get_series(revision=wantver, sloppytrailers=cmdargs.sloppytrailers, reroll=reroll,
                           codereview_trailers=load_codereview)
    if lser is None and cmdargs.cherrypick != '_':
        if wantver is None:
            logger.critical('No patches found.')
        else:
            logger.critical('Unable to find revision %s', wantver)
        return
    if lser is None:
        logger.critical('No patches found.')
        return

    if len(lmbx.series) > 1 and not wantver:
        logger.info('Will use the latest revision: v%s', lser.revision)
        logger.info('You can pick other revisions using the -vN flag')

    if cmdargs.cherrypick:
        cherrypick = list()
        if cmdargs.cherrypick == '_':
            # We might want to pick a patch sent as a followup, so create a fake series
            # and add followups with diffs
            if lser is None:
                lser = b4.LoreSeries(revision=1, expected=1)
            for followup in lmbx.followups:
                if followup.has_diff:
                    lser.add_patch(followup)
            # Only grab the exact msgid provided
            for at, lmsg in enumerate(lser.patches[1:], 1):
                if lmsg and lmsg.msgid == msgid:
                    cherrypick = [at]
                    cmdargs.cherrypick = f'<{msgid}>'
                    break
            if not len(cherrypick):
                logger.critical('Specified msgid is not present in the series, cannot cherrypick')
                sys.exit(1)
        elif cmdargs.cherrypick.find('*') >= 0:
            # Globbing on subject
            at = 0
            for lmsg in lser.patches[1:]:
                at += 1
                if lmsg is None:
                    continue
                if fnmatch.fnmatch(lmsg.subject, cmdargs.cherrypick):
                    cherrypick.append(at)
            if not len(cherrypick):
                logger.critical('Could not match "%s" to any subjects in the series', cmdargs.cherrypick)
                sys.exit(1)
        else:
            cherrypick = list(b4.parse_int_range(cmdargs.cherrypick, upper=len(lser.patches) - 1))
    else:
        cherrypick = None

    am_msgs = lser.get_am_ready(noaddtrailers=cmdargs.noaddtrailers, addmysob=cmdargs.addmysob, addlink=cmdargs.addlink,
                                cherrypick=cherrypick, copyccs=cmdargs.copyccs, allowbadchars=cmdargs.allowbadchars,
                                showchecks=cmdargs.check)
    logger.info('---')

    if cherrypick is None:
        logger.critical('Total patches: %s', len(am_msgs))
    else:
        logger.info('Total patches: %s (cherrypicked: %s)', len(am_msgs), cmdargs.cherrypick)

    if len(lser.trailer_mismatches):
        logger.critical('---')
        logger.critical('NOTE: some trailers ignored due to from/email mismatches:')
        for tname, tvalue, fname, femail in lser.trailer_mismatches:
            logger.critical('    ! Trailer: %s: %s', tname, tvalue)
            logger.critical('     Msg From: %s <%s>', fname, femail)
        logger.critical('NOTE: Rerun with -S to apply them anyway')

    top_msgid = None
    first_body = None
    for lmsg in lser.patches:
        if lmsg is not None:
            first_body = lmsg.body
            top_msgid = lmsg.msgid
            break
    if top_msgid is None or first_body is None:
        logger.critical('Could not find any patches in the series.')
        return

    topdir = b4.git_get_toplevel()

    if cmdargs.threeway:
        if not topdir:
            logger.critical('WARNING: cannot prepare 3-way (not in a git dir)')
        elif not lser.complete:
            logger.critical('WARNING: cannot prepare 3-way (series incomplete)')
        else:
            _checked, mismatches = lser.check_applies_clean(gitdir=topdir)
            if mismatches:
                rstart, rend = lser.make_fake_am_range(gitdir=None)
                if rstart and rend:
                    logger.info('Prepared fake commit range for 3-way merge (%.12s..%.12s)', rstart, rend)

    logger.critical('---')
    if lser.partial_reroll:
        logger.critical('WARNING: v%s is a partial reroll from previous revisions', lser.revision)
        logger.critical('         Please carefully review the resulting series to ensure correctness')
        logger.critical('         Pass --no-partial-reroll to disable')
        logger.critical('---')
    if not lser.complete and not cmdargs.cherrypick:
        logger.critical('WARNING: Thread incomplete!')

    am_filename = None

    linkmask = str(config.get('linkmask', ''))
    if '%s' not in linkmask:
        logger.critical('ERROR: linkmask must contain %s for the message-id')
        sys.exit(1)

    if cmdargs.subcmd == 'am':
        wantname = cmdargs.wantname
        if cmdargs.maildir or config.get('save-maildirs', 'no') == 'yes':
            save_maildir = True
            dftext = 'maildir'
        else:
            save_maildir = False
            dftext = 'mbx'

        if wantname:
            slug = wantname
            if wantname.find('.') > -1:
                slug = '.'.join(wantname.split('.')[:-1])
        else:
            slug = lser.get_slug(extended=True)

        if outdir != '-':
            pathlib.Path(outdir).mkdir(parents=True, exist_ok=True)

            am_filename = os.path.join(outdir, f'{slug}.{dftext}')
            am_cover = os.path.join(outdir, f'{slug}.cover')

            if os.path.exists(am_filename):
                if os.path.isdir(am_filename):
                    shutil.rmtree(am_filename)
                else:
                    os.unlink(am_filename)
            if save_maildir:
                b4.save_maildir(am_msgs, am_filename)
            else:
                with open(am_filename, 'wb') as fh:
                    b4.save_git_am_mbox(am_msgs, fh)
        else:
            am_cover = None
            b4.save_git_am_mbox(am_msgs, sys.stdout.buffer)

        if am_cover and lser.has_cover and not cmdargs.nocover:
            lser.save_cover(am_cover)

        linkurl = linkmask % top_msgid
        if cmdargs.quiltready:
            q_dirname = os.path.join(outdir, f'{slug}.patches')
            save_as_quilt(am_msgs, q_dirname)
            logger.critical('Quilt: %s', q_dirname)

        logger.critical(' Link: %s', linkurl)

    if cmdargs.subcmd == 'shazam':
        if not topdir:
            logger.critical('Could not figure out where your git dir is, cannot shazam.')
            sys.exit(1)

        ifh = io.BytesIO()
        if lser.prereq_patch_ids:
            logger.info(' Deps: looking for dependencies matching %s patch-ids', len(lser.prereq_patch_ids))
            query = ' OR '.join([f'patchid:{x}' for x in lser.prereq_patch_ids])
            logger.debug('query=%s', query)
            dmsgs = b4.get_pi_search_results(query)
            pmap = dict()
            if dmsgs:
                for dmsg in dmsgs:
                    dbody, _ = b4.LoreMessage.get_payload(dmsg)
                    if not b4.DIFF_RE.search(dbody):
                        continue
                    dlmsg = b4.LoreMessage(dmsg)
                    if dlmsg.git_patch_id in lser.prereq_patch_ids:
                        logger.debug('%s => %s', dlmsg.git_patch_id, dlmsg.subject)
                        pmap[dlmsg.git_patch_id] = dlmsg
            for ppid in lser.prereq_patch_ids:
                if ppid in pmap:
                    logger.info(' Deps: Applying prerequisite patch: %s', pmap[ppid].full_subject)
                    pam_msg = pmap[ppid].get_am_message(add_trailers=False)
                    b4.save_mboxrd_mbox([pam_msg], ifh)

        b4.save_git_am_mbox(am_msgs, ifh)
        ambytes = ifh.getvalue()
        if not cmdargs.makefetchhead:
            amflags = str(config.get('shazam-am-flags', ''))
            sp = shlex.shlex(amflags, posix=True)
            sp.whitespace_split = True
            amargs = list(sp) + ['--patch-format=mboxrd']
            ecode, out = b4.git_run_command(topdir, ['am'] + amargs, stdin=ambytes, logstderr=True, rundir=topdir)
            logger.info(out.strip())
            if ecode == 0:
                thanks_record_am(lser, cherrypick=cherrypick)
            sys.exit(ecode)

        base_commit = get_base_commit(topdir, first_body, lser, cmdargs)
        linkurl = linkmask % top_msgid

        merge_template = DEFAULT_MERGE_TEMPLATE
        if config.get('shazam-merge-template'):
            # Try to load this template instead
            try:
                merge_template = b4.read_template(str(config['shazam-merge-template']))
            except FileNotFoundError:
                logger.critical('ERROR: shazam-merge-template says to use %s, but it does not exist',
                                config['shazam-merge-template'])
                sys.exit(2)

        if lser.has_cover and lser.patches[0] is not None:
            clmsg: b4.LoreMessage = lser.patches[0]
            parts = b4.LoreMessage.get_body_parts(clmsg.body)
            covermessage = parts[1]
        else:
            if lser.patches[1] is None:
                logger.critical('No cover letter provided by the author and no first patch, cannot shazam')
                sys.exit(1)

            clmsg = lser.patches[1]
            covermessage = ('NOTE: No cover letter provided by the author.\n'
                            '      Add merge commit message here.')

        tptvals = {
            'seriestitle': clmsg.subject,
            'authorname': clmsg.fromname,
            'authoremail': clmsg.fromemail,
            'covermessage': covermessage,
            'mid': top_msgid,
            'midurl': linkurl,
        }
        if len(am_msgs) > 1:
            tptvals['patch_or_series'] = 'patch series'
        else:
            tptvals['patch_or_series'] = 'patch'

        mergeflags = str(config.get('shazam-merge-flags', '--signoff'))

        am_flags = ['-3']
        amflags_cfg = str(config.get('shazam-am-flags', ''))
        if amflags_cfg:
            sp = shlex.shlex(amflags_cfg, posix=True)
            sp.whitespace_split = True
            am_flags.extend(list(sp))

        try:
            if cmdargs.mergebase:
                logger.info(' Base: %s', base_commit)
            else:
                logger.info(' Base: %s (use --merge-base to override)', base_commit)
            b4.git_fetch_am_into_repo(topdir, ambytes=ambytes, at_base=base_commit,
                                       origin=linkurl, am_flags=am_flags)
        except b4.AmConflictError as cex:
            gwt = cex.worktree_path
            if not getattr(cmdargs, 'shazam_resolve', False):
                b4.git_run_command(topdir, ['worktree', 'remove', '--force', gwt])
                logger.critical('Unable to cleanly apply series, see failure log below')
                logger.critical('---')
                logger.critical(cex.output)
                logger.critical('---')
                logger.critical('Not fetching into FETCH_HEAD')
                logger.critical('Use --resolve to enable conflict resolution')
                sys.exit(1)

            common_dir = b4.git_get_common_dir(topdir)
            if not common_dir:
                logger.critical('Unable to determine git common dir')
                b4.git_run_command(topdir, ['worktree', 'remove', '--force', gwt])
                sys.exit(1)

            state = {
                'origin': linkurl,
                'merge_template_values': tptvals,
                'merge_template': merge_template,
                'merge_flags': mergeflags,
                'no_interactive': cmdargs.no_interactive,
            }
            _start_merge_resolve(topdir, cex, common_dir, state)
        except RuntimeError:
            sys.exit(1)

        gitargs = ['rev-parse', '--git-dir']
        ecode, out = b4.git_run_command(topdir, gitargs, logstderr=True)
        if ecode > 0:
            logger.critical('Unable to find git directory')
            logger.critical(out.strip())
            sys.exit(ecode)
        mmf = os.path.join(out.rstrip(), 'b4-cover')

        # Write out a sample merge message using the cover letter
        if os.path.exists(mmf):
            # Make sure any old cover letters don't confuse anyone
            os.unlink(mmf)

        body = Template(merge_template).safe_substitute(tptvals)
        with open(mmf, 'w') as mmh:
            mmh.write(body)

        sp = shlex.shlex(mergeflags, posix=True)
        sp.whitespace_split = True
        if cmdargs.no_interactive:
            edit = '--no-edit'
        else:
            edit = '--edit'
        mergeargs = ['merge', '--no-ff', '-F', mmf, edit, 'FETCH_HEAD'] + list(sp)
        mergecmd = ['git'] + mergeargs

        thanks_record_am(lser, cherrypick=cherrypick)
        if cmdargs.merge:
            if not cmdargs.no_interactive:
                logger.info('Will exec: %s', ' '.join(mergecmd))
                try:
                    input('Press Enter to continue or Ctrl-C to abort')
                except KeyboardInterrupt:
                    logger.info('')
                    sys.exit(130)
            else:
                logger.info('Invoking: %s', ' '.join(mergecmd))
            if hasattr(sys, '_running_in_pytest'):
                # Don't execvp, as this kills our tests
                _out = b4.git_run_command(None, mergeargs)
                sys.exit(_out[0])

            # We exec git-merge and let it take over
            os.execvp(mergecmd[0], mergecmd)

        logger.info('You can now merge or checkout FETCH_HEAD')
        logger.info('  e.g.: %s', ' '.join(mergecmd))
        sys.exit(0)

    if topdir:
        base_commit = get_base_commit(topdir, first_body, lser, cmdargs)
        checked, mismatched = lser.check_applies_clean(topdir, at=cmdargs.guessbranch)
        if checked and len(mismatched) == 0 and checked != len(mismatched):
            logger.critical(' Base: applies clean to current tree')
        elif base_commit != 'HEAD':
            gitbranch = lser.get_slug(extended=False)
            logger.critical('       git checkout -b %s %s', gitbranch, base_commit)

    if cmdargs.outdir != '-':
        logger.critical('       git am %s%s', '-3 ' if cmdargs.threeway else '', am_filename)

    thanks_record_am(lser, cherrypick=cherrypick)


def thanks_record_am(lser: b4.LoreSeries, cherrypick: Optional[List[int]]) -> None:
    # Are we tracking this already?
    datadir = b4.get_data_dir()
    slug = lser.get_slug(extended=True)
    filename = '%s.am' % slug

    patches = list()
    msgids = list()
    at = 0
    padlen = len(str(lser.expected))
    lmsg = None

    for pmsg in lser.patches:
        if pmsg is None:
            at += 1
            continue

        if lmsg is None:
            lmsg = pmsg

        if not pmsg.has_diff:
            # Don't care about the cover letter
            at += 1
            continue

        if cherrypick is not None and at not in cherrypick:
            logger.debug('Skipped non-cherrypicked: %s', at)
            at += 1
            continue

        # Add it for patchwork state tracking
        msgids.append(pmsg.msgid)

        if pmsg.pwhash is None:
            logger.debug('Unable to get hashes for all patches, not tracking for thanks')
            return

        prefix = '%s/%s' % (str(pmsg.counter).zfill(padlen), pmsg.expected)
        patches.append((pmsg.subject, pmsg.pwhash, pmsg.msgid, prefix))
        at += 1

    if lmsg is None:
        logger.debug('All patches missing, not tracking for thanks')
        return

    try:
        allto = email.utils.getaddresses([str(x) for x in lmsg.msg.get_all('to', [])])
    except Exception as ex:
        allto = []
        logger.debug('Unable to parse the To: header in %s: %s', lmsg.msgid, str(ex))
    try:
        allcc = email.utils.getaddresses([str(x) for x in lmsg.msg.get_all('cc', [])])
    except Exception as ex:
        allcc = []
        logger.debug('Unable to parse the Cc: header in %s: %s', lmsg.msgid, str(ex))

    # TODO: check for reply-to and x-original-from
    out = {
        'msgid': lmsg.msgid,
        'subject': lmsg.full_subject,
        'fromname': lmsg.fromname,
        'fromemail': lmsg.fromemail,
        'to': b4.format_addrs(allto, clean=False),
        'cc': b4.format_addrs(allcc, clean=False),
        'references': b4.LoreMessage.clean_header(lmsg.msg['References']),
        'sentdate': b4.LoreMessage.clean_header(lmsg.msg['Date']),
        'quote': b4.make_quote(lmsg.body, maxlines=5),
        'cherrypick': cherrypick is not None,
        'patches': patches,
    }
    fullpath = os.path.join(datadir, filename)
    with open(fullpath, 'w', encoding='utf-8') as fh:
        json.dump(out, fh, ensure_ascii=False, indent=4)
        logger.debug('Wrote %s for thanks tracking', filename)

    config = b4.get_main_config()
    pwstate = str(config.get('pw-review-state', ''))
    if pwstate:
        b4.patchwork_set_state(msgids, pwstate)


def save_as_quilt(am_msgs: List[EmailMessage], q_dirname: str) -> None:
    if os.path.exists(q_dirname):
        logger.critical('ERROR: Directory %s exists, not saving quilt patches', q_dirname)
        return
    pathlib.Path(q_dirname).mkdir(parents=True)
    patch_filenames = list()
    for msg in am_msgs:
        lsubj = b4.LoreSubject(msg.get('subject', ''))
        slug = '%04d_%s' % (lsubj.counter, re.sub(r'\W+', '_', lsubj.subject).strip('_').lower())
        patch_filename = f'{slug}.patch'
        patch_filenames.append(patch_filename)
        quilt_out = os.path.join(q_dirname, patch_filename)
        i, m, p = b4.get_mailinfo(msg.as_bytes(policy=b4.emlpolicy), scissors=True)
        with open(quilt_out, 'wb') as fh:
            if i.get('Author'):
                fh.write(b'From: %s <%s>\n' % (i.get('Author', '').encode(), i.get('Email', '').encode()))
            else:
                fh.write(b'From: %s\n' % i.get('Email', '').encode())
            fh.write(b'Subject: %s\n' % i.get('Subject', '').encode())
            fh.write(b'Date: %s\n' % i.get('Date', '').encode())
            fh.write(b'\n')
            fh.write(m)
            fh.write(p)
        logger.debug('  Wrote: %s', patch_filename)
    # Write the series file
    with open(os.path.join(q_dirname, 'series'), 'w') as sfh:
        for patch_filename in patch_filenames:
            sfh.write('%s\n' % patch_filename)


def get_extra_series(msgs: List[EmailMessage], direction: int = 1, wantvers: Optional[List[int]] = None,
                     nocache: bool = False) -> List[EmailMessage]:
    base_msg: Optional[EmailMessage] = None
    latest_revision: Optional[int] = None
    seen_msgids: Set[str] = set()
    seen_covers: Set[int] = set()
    queries: Set[str] = set()
    change_ids: Set[str] = set()
    by_changeid: bool = False
    for msg in msgs:
        msgid = b4.LoreMessage.get_clean_msgid(msg)
        if msgid is None:
            continue
        seen_msgids.add(msgid)
        lsub = b4.LoreSubject(msg['Subject'])

        # Ignore patches above 1
        if lsub.counter > 1:
            continue

        if base_msg is not None:
            logger.debug('Current base_msg: %s', base_msg['Subject'])

        if lsub.reply:
            logger.debug('Ignoring reply: %s', lsub.full_subject)
            continue

        payload, _ = b4.LoreMessage.get_payload(msg)
        if payload:
            matches = re.search(r'^change-id:\s+(\S+)', payload, flags=re.I | re.M)
            if matches:
                cid = matches.groups()[0]
                logger.debug('Found change-id %s', cid)
                change_ids.add(cid)
                by_changeid = True
                q = 'nq:"change-id: %s"' % cid
                queries.add(q)

        logger.debug('Checking the subject on %s', lsub.full_subject)
        if latest_revision is None or lsub.revision >= latest_revision:
            latest_revision = lsub.revision
            if lsub.counter == 0 and not lsub.counters_inferred:
                # And a cover letter, nice. This is the easy case
                base_msg = msg
                seen_covers.add(latest_revision)
            elif lsub.counter == 1 and latest_revision not in seen_covers:
                # A patch/series without a cover letter
                base_msg = msg

    if base_msg is None or latest_revision is None:
        return msgs

    # Get subject info from base_msg again
    lsub = b4.LoreSubject(base_msg['Subject'])
    if not len(lsub.prefixes):
        logger.debug('Not checking for new revisions: no prefixes on the cover letter.')
        return msgs
    if direction < 0 and latest_revision <= 1:
        logger.debug('This is the earliest version of the series')
        return msgs
    if direction < 0 and wantvers is None:
        wantvers = [latest_revision - 1]

    msgdate = email.utils.parsedate_tz(str(base_msg['Date']))
    if msgdate is None:
        logger.debug('Unable to parse the date, not checking for revisions')
        return msgs

    if not by_changeid:
        logger.debug('Using subject+from combo for query')
        fromeml = email.utils.getaddresses(base_msg.get_all('from', []))[0][1]
        q = '(s:"%s" AND f:"%s")' % (lsub.subject.replace('"', ''), fromeml)
        queries.add(q)

    startdate = time.strftime('%Y%m%d', msgdate[:9])
    if direction > 0:
        logger.critical('Checking for newer revisions')
        datelim = 'd:%s..' % startdate
    else:
        logger.critical('Checking for older revisions')
        # Cap backward search to 12 months to avoid matching years of
        # identically-named series (common with subject+from fallback).
        earliest = time.strftime('%Y%m%d', time.gmtime(
            time.mktime(msgdate[:9]) - 365 * 86400))
        datelim = 'd:%s..%s' % (earliest, startdate)

    q = '(%s) AND %s' % (' OR '.join(queries), datelim)
    logger.debug('Query: %s', q)
    q_msgs = b4.get_pi_search_results(q, nocache=nocache)
    if not q_msgs:
        return msgs

    seen_revisions: Dict[int, int] = dict()
    candidates: List[EmailMessage] = list()
    for q_msg in q_msgs:
        q_msgid = b4.LoreMessage.get_clean_msgid(q_msg)
        if q_msgid is None:
            continue
        _subject = q_msg.get('Subject', '(no subject)')
        lsub = b4.LoreSubject(_subject)
        if q_msgid in seen_msgids:
            logger.debug('Skipping %s: already have it', lsub.full_subject)
            continue
        if lsub.reply:
            # These will get sorted out later
            logger.debug('Adding reply: %s', lsub.full_subject)
            logger.debug('              msgid: %s', q_msgid)
            candidates.append(q_msg)
            seen_msgids.add(q_msgid)
            continue

        if direction > 0 and lsub.revision <= latest_revision:
            logger.debug('Ignoring result (not new revision): %s', lsub.full_subject)
            continue
        if direction < 0 and lsub.revision >= latest_revision:
            logger.debug('Ignoring result (not old revision): %s', lsub.full_subject)
            continue
        if direction < 0 and wantvers and lsub.revision not in wantvers:
            logger.debug('Ignoring result (not revision we want): %s', lsub.full_subject)
            continue

        if lsub.revision == 1 and lsub.revision == latest_revision:
            # Someone sent a separate message with an identical title but no new vX in the subject line
            if direction > 0:
                # It's *probably* a new revision.
                logger.debug('Likely a new revision: %s', lsub.full_subject)
            else:
                # It's *probably* an older revision.
                logger.debug('Likely an older revision: %s', lsub.full_subject)
        elif direction > 0 and lsub.revision > latest_revision:
            logger.debug('Definitely a new revision [v%s]: %s', lsub.revision, lsub.full_subject)
        elif direction < 0 and lsub.revision < latest_revision:
            logger.debug('Definitely an older revision [v%s]: %s', lsub.revision, lsub.full_subject)
        else:
            logger.debug('No idea what this is: %s', lsub.subject)
            continue
        if lsub.revision not in seen_revisions:
            seen_revisions[lsub.revision] = 0
        seen_revisions[lsub.revision] += 1
        logger.debug('Adding: %s', lsub.full_subject)
        candidates.append(q_msg)
        seen_msgids.add(q_msgid)

    if by_changeid and change_ids:
        # Verify that candidate revisions actually belong to the same
        # series.  The nq: search can match messages that merely
        # reference our change-id as a prerequisite-change-id, so we
        # check that at least one message per revision contains a
        # standalone "change-id: <id>" line (not prerequisite-change-id).
        valid_revisions: Set[int] = set()
        for q_msg in candidates:
            payload, _ = b4.LoreMessage.get_payload(q_msg)
            if not payload:
                continue
            for cid in change_ids:
                if re.search(rf'^change-id:\s*{re.escape(cid)}\s*$', payload, flags=re.I | re.M):
                    lsub = b4.LoreSubject(q_msg.get('Subject', ''))
                    valid_revisions.add(lsub.revision)
                    break
        rejected: List[int] = list()
        validated: List[EmailMessage] = list()
        for q_msg in candidates:
            lsub = b4.LoreSubject(q_msg.get('Subject', ''))
            if lsub.reply or lsub.revision in valid_revisions:
                validated.append(q_msg)
            else:
                if lsub.revision not in rejected:
                    rejected.append(lsub.revision)
                if lsub.revision in seen_revisions:
                    seen_revisions[lsub.revision] -= 1
        candidates = validated
        for rev in rejected:
            logger.debug('Rejected v%s: change-id mismatch', rev)
        # Clean up empty revision counts
        seen_revisions = {r: c for r, c in seen_revisions.items() if c > 0}

    msgs.extend(candidates)

    for rev, count in seen_revisions.items():
        logger.info('  Added from v%s: %s patches', rev, count)

    return msgs


def refetch(dest: str) -> None:
    mbox: Union[mailbox.Maildir, mailbox.mbox]
    if b4.is_maildir(dest):
        mbox = mailbox.Maildir(dest)
    else:
        mbox = mailbox.mbox(dest)

    by_msgid: Dict[str, EmailMessage] = dict()
    for key, msg in mbox.items():
        # We normally pass EmailMessage objects, but this works, too
        msgid = b4.LoreMessage.get_clean_msgid(msg)  # type: ignore[arg-type]
        if not msgid:
            continue
        if msgid not in by_msgid:
            amsgs = b4.get_pi_thread_by_msgid(msgid, nocache=True)
            if amsgs:
                for amsg in amsgs:
                    amsgid = b4.LoreMessage.get_clean_msgid(amsg)
                    if not amsgid:
                        continue
                    if amsgid not in by_msgid:
                        by_msgid[amsgid] = amsg
        if msgid in by_msgid:
            mbox.update(((key, by_msgid[msgid]),))
            logger.info('Refetched: %s', msg.get('Subject'))
        else:
            logger.warning('WARNING: Message-id not known: %s', msgid)
    mbox.close()


def minimize_thread(msgs: List[EmailMessage]) -> List[EmailMessage]:
    # We go through each message and minimize headers and body content
    wanthdrs = {
                'From',
                'Subject',
                'Date',
                'Message-ID',
                'Reply-To',
                'In-Reply-To',
                }
    mmsgs = list()
    for msg in msgs:
        mmsg = EmailMessage()
        for wanthdr in wanthdrs:
            cleanhdr = b4.LoreMessage.clean_header(msg[wanthdr])
            if cleanhdr:
                mmsg[wanthdr] = cleanhdr

        body, _charset = b4.LoreMessage.get_payload(msg)
        if not (b4.DIFF_RE.search(body) or b4.DIFFSTAT_RE.search(body)):
            _htrs, cmsg, _mtrs, _basement, _sig = b4.LoreMessage.get_body_parts(body)
            # split the message into quoted and unquoted chunks
            chunks: List[Tuple[bool, List[str]]] = list()
            chunk: List[str] = list()
            current = None
            for line in (cmsg.rstrip().splitlines()):
                quoted = line.startswith('>') and True or False
                if current is None:
                    current = quoted
                if current == quoted:
                    if quoted and re.search(r'^>\s*>', line):
                        # trim multiple levels of quoting
                        continue
                    if quoted and not chunk and line.strip() == '>':
                        # Trim empty lines with just > in them
                        continue
                    chunk.append(line)
                    continue

                if current:
                    while len(chunk) and chunk[-1].strip() == '>':
                        chunk.pop(-1)
                if chunk:
                    chunks.append((quoted, chunk))
                chunk = list()
                if not (quoted and line.strip() == '>'):
                    chunk.append(line)
                current = quoted

            if current is None:
                current = False

            # Don't append bottom quotes
            if chunk and not current:
                chunks.append((current, chunk))

            body = ''
            for _quoted, chunk in chunks:
                # Should we offer a way to trim the quote in some fashion?
                body += '\n'.join(chunk).strip() + '\n\n'
            if not body.strip():
                continue

        mmsg.set_payload(body, charset='utf-8')
        # mmsg.set_charset('utf-8')
        mmsgs.append(mmsg)

    return mmsgs


def _start_merge_resolve(topdir: str, cex: b4.AmConflictError,
                          common_dir: str, state: Dict[str, Any]) -> None:
    gwt = cex.worktree_path
    logger.critical('---')
    logger.critical(cex.output)
    logger.critical('---')
    logger.critical('Patch series did not apply cleanly, resolving...')

    # Find rebase-apply in the worktree
    ecode, gitdir = b4.git_run_command(gwt, ['rev-parse', '--git-dir'],
                                        logstderr=True, rundir=gwt)
    if ecode > 0:
        logger.critical('Unable to find git directory in worktree')
        b4.git_run_command(topdir, ['worktree', 'remove', '--force', gwt])
        sys.exit(1)
    rebase_apply = os.path.join(gitdir.strip(), 'rebase-apply')
    if not os.path.isdir(rebase_apply):
        logger.critical('No git-am state found in worktree.')
        b4.git_run_command(topdir, ['worktree', 'remove', '--force', gwt])
        sys.exit(1)

    # Extract remaining patches
    with open(os.path.join(rebase_apply, 'next'), 'r') as fh:
        next_num = int(fh.read().strip())
    with open(os.path.join(rebase_apply, 'last'), 'r') as fh:
        last_num = int(fh.read().strip())

    patches_dir = os.path.join(common_dir, 'b4-shazam-patches')
    if os.path.exists(patches_dir):
        shutil.rmtree(patches_dir)
    os.makedirs(patches_dir)

    patch_count = 0
    for i in range(next_num, last_num + 1):
        src = os.path.join(rebase_apply, f'{i:04d}')
        if os.path.exists(src):
            dst = os.path.join(patches_dir, f'{patch_count:04d}')
            shutil.copy2(src, dst)
            patch_count += 1

    with open(os.path.join(patches_dir, 'total'), 'w') as fh:
        fh.write(str(patch_count))
    with open(os.path.join(patches_dir, 'current'), 'w') as fh:
        fh.write('0')

    # Check for uncommitted changes
    status_lines = b4.git_get_repo_status(topdir)
    if status_lines:
        logger.critical('You have uncommitted changes in your working tree.')
        logger.critical('Please commit or stash them before resolving.')
        shutil.rmtree(patches_dir)
        b4.git_run_command(topdir, ['worktree', 'remove', '--force', gwt])
        sys.exit(1)

    # Fetch successfully applied patches into FETCH_HEAD
    logger.info('Fetching successfully applied patches into FETCH_HEAD')
    ecode, out = b4.git_run_command(topdir, ['fetch', gwt], logstderr=True)
    if ecode > 0:
        logger.critical('Unable to fetch from the worktree')
        logger.critical(out.strip())
        shutil.rmtree(patches_dir)
        b4.git_run_command(topdir, ['worktree', 'remove', '--force', gwt])
        sys.exit(1)

    # Rewrite FETCH_HEAD origin
    origin = state.get('origin')
    if origin:
        gitargs = ['rev-parse', '--git-path', 'FETCH_HEAD']
        ecode, fhf = b4.git_run_command(topdir, gitargs, logstderr=True)
        if ecode == 0:
            fhf = fhf.rstrip()
            with open(fhf, 'r') as fhh:
                contents = fhh.read()
            mmsg = 'patches from %s' % origin
            new_contents = contents.replace(gwt, mmsg)
            if new_contents != contents:
                with open(fhf, 'w') as fhh:
                    fhh.write(new_contents)

    # Remove the worktree
    b4.git_run_command(topdir, ['worktree', 'remove', '--force', gwt])

    # Save state for --continue/--abort
    state_file = os.path.join(common_dir, 'b4-shazam-state.json')
    with open(state_file, 'w') as sfh:
        json.dump(state, sfh, indent=2)

    # Start merge of successfully applied patches
    logger.info('Merging successfully applied patches into your branch...')
    ecode, out = b4.git_run_command(topdir, ['merge', '--no-ff', '--no-commit', 'FETCH_HEAD'],
                                     logstderr=True, rundir=topdir)

    if ecode > 0:
        logger.warning('Merge had conflicts:')
        logger.warning(out.strip())
        logger.warning('Resolve conflicts, then run: b4 shazam --continue')
        logger.warning('To abort: b4 shazam --abort')
        sys.exit(1)

    # Merge was clean, apply remaining patches
    _apply_remaining_patches(topdir, patches_dir, state, state_file, common_dir)
    sys.exit(0)


def _apply_remaining_patches(topdir: str, patches_dir: str, state: Dict[str, Any],
                              state_file: str, common_dir: str) -> None:
    with open(os.path.join(patches_dir, 'total'), 'r') as fh:
        total = int(fh.read().strip())
    with open(os.path.join(patches_dir, 'current'), 'r') as fh:
        current = int(fh.read().strip())

    while current < total:
        patch_file = os.path.join(patches_dir, f'{current:04d}')
        if not os.path.exists(patch_file):
            current += 1
            continue

        with open(patch_file, 'rb') as fh:
            patch_data = fh.read()

        logger.info('Applying remaining patch %d/%d...', current + 1, total)
        ecode, out = b4.git_run_command(topdir, ['apply', '--3way'],
                                         stdin=patch_data, logstderr=True, rundir=topdir)
        if ecode > 0:
            logger.critical('---')
            logger.critical(out.strip())
            logger.critical('---')
            logger.critical('Remaining patch %d/%d did not apply cleanly.', current + 1, total)
            logger.critical('Resolve conflicts in your working tree, then run: b4 shazam --continue')
            logger.critical('To abort: b4 shazam --abort')
            # Advance past this patch, its changes (with conflict markers) are in the tree
            with open(os.path.join(patches_dir, 'current'), 'w') as fh:
                fh.write(str(current + 1))
            sys.exit(1)

        # Patch applied cleanly, stage it
        b4.git_run_command(topdir, ['add', '-u'], logstderr=True, rundir=topdir)
        current += 1
        with open(os.path.join(patches_dir, 'current'), 'w') as fh:
            fh.write(str(current))

    # All patches applied, finish the merge
    _finish_shazam_merge(topdir, state, state_file, common_dir, patches_dir)


def _finish_shazam_merge(topdir: str, state: Dict[str, Any], state_file: str,
                          common_dir: str, patches_dir: str) -> None:
    b4.git_run_command(topdir, ['add', '-u'], logstderr=True, rundir=topdir)

    gitargs = ['rev-parse', '--git-dir']
    ecode, out = b4.git_run_command(topdir, gitargs, logstderr=True)
    if ecode > 0:
        logger.critical('Unable to find git directory')
        sys.exit(1)
    mmf = os.path.join(out.rstrip(), 'b4-cover')

    merge_template = state.get('merge_template', DEFAULT_MERGE_TEMPLATE)
    tptvals = state.get('merge_template_values', {})

    body = Template(merge_template).safe_substitute(tptvals)
    with open(mmf, 'w') as mmh:
        mmh.write(body)

    # Clean up state before committing -- if the commit is interactive
    # (execvp), we won't get a chance to clean up after.
    if os.path.exists(patches_dir):
        shutil.rmtree(patches_dir)
    if os.path.exists(state_file):
        os.unlink(state_file)

    no_interactive = state.get('no_interactive', False)
    mergeflags = str(state.get('merge_flags', ''))
    commitargs = ['commit', '-F', mmf]
    if mergeflags:
        sp = shlex.shlex(mergeflags, posix=True)
        sp.whitespace_split = True
        commitargs.extend(list(sp))
    if no_interactive:
        commitargs.append('--no-edit')
        ecode, out = b4.git_run_command(topdir, commitargs, logstderr=True, rundir=topdir)
        if ecode > 0:
            logger.critical('Failed to commit merge:')
            logger.critical(out.strip())
            sys.exit(1)
        logger.info(out.strip())
    else:
        # Interactive, need the terminal, so exec git directly
        commitargs.append('--edit')
        commitcmd = ['git'] + commitargs
        logger.info('Invoking: %s', ' '.join(commitcmd))
        if hasattr(sys, '_running_in_pytest'):
            _out = b4.git_run_command(None, commitargs)
            sys.exit(_out[0])
        os.chdir(topdir)
        os.execvp(commitcmd[0], commitcmd)

    if os.path.exists(mmf):
        os.unlink(mmf)
    logger.info('Merge completed successfully.')


def _load_shazam_state(require_state: bool = True) -> Tuple[str, str, str, Optional[Dict[str, Any]]]:
    topdir = b4.git_get_toplevel()
    if not topdir:
        logger.critical('Could not figure out where your git dir is.')
        sys.exit(1)
    common_dir = b4.git_get_common_dir(topdir)
    if not common_dir:
        logger.critical('Unable to determine git common dir.')
        sys.exit(1)

    state_file = os.path.join(common_dir, 'b4-shazam-state.json')
    state = None
    if require_state:
        if not os.path.exists(state_file):
            logger.critical('No shazam state found. Nothing to continue.')
            sys.exit(1)
        with open(state_file, 'r') as fh:
            state = json.load(fh)
        patches_dir = os.path.join(common_dir, 'b4-shazam-patches')
        if not os.path.isdir(patches_dir):
            logger.critical('Patches directory not found. State may be corrupted.')
            logger.critical('Run: b4 shazam --abort')
            sys.exit(1)

    return topdir, common_dir, state_file, state


def shazam_continue(cmdargs: argparse.Namespace) -> None:
    topdir, common_dir, state_file, state = _load_shazam_state(require_state=True)
    assert state is not None
    patches_dir = os.path.join(common_dir, 'b4-shazam-patches')

    # Stage any resolved files
    b4.git_run_command(topdir, ['add', '-u'], logstderr=True, rundir=topdir)

    # Check for remaining unmerged files
    _ecode, unmerged = b4.git_run_command(topdir, ['diff', '--name-only', '--diff-filter=U'],
                                          logstderr=True, rundir=topdir)
    if unmerged.strip():
        logger.critical('There are still unresolved conflicts:')
        logger.critical(unmerged.strip())
        logger.critical('Resolve them, then run: b4 shazam --continue')
        sys.exit(1)

    # Apply remaining patches and finish merge
    _apply_remaining_patches(topdir, patches_dir, state, state_file, common_dir)


def shazam_abort(cmdargs: argparse.Namespace) -> None:
    topdir, common_dir, state_file, _state = _load_shazam_state(require_state=False)
    found = False

    # Abort in-progress merge if any
    b4.git_run_command(topdir, ['merge', '--abort'], logstderr=True, rundir=topdir)

    # Clean up patches directory
    patches_dir = os.path.join(common_dir, 'b4-shazam-patches')
    if os.path.exists(patches_dir):
        shutil.rmtree(patches_dir)
        found = True

    # Clean up worktree if it exists
    gwt = os.path.join(common_dir, 'b4-shazam-worktree')
    if os.path.exists(gwt):
        b4.git_run_command(topdir, ['worktree', 'remove', '--force', gwt])
        found = True

    if os.path.exists(state_file):
        os.unlink(state_file)
        found = True

    if found:
        logger.info('Shazam aborted and cleaned up.')
    else:
        logger.info('No shazam in progress.')


def main(cmdargs: argparse.Namespace) -> None:
    # We force some settings
    if cmdargs.subcmd == 'shazam':
        if getattr(cmdargs, 'shazam_continue', False):
            return shazam_continue(cmdargs)
        if getattr(cmdargs, 'shazam_abort', False):
            return shazam_abort(cmdargs)
        cmdargs.checknewer = True
        cmdargs.threeway = False
        cmdargs.nopartialreroll = False
        cmdargs.outdir = '-'
        cmdargs.guessbranch = None
        if cmdargs.merge:
            cmdargs.makefetchhead = True
        if cmdargs.makefetchhead:
            cmdargs.guessbase = True
        else:
            cmdargs.guessbase = False
    else:
        cmdargs.mergebase = False

    if cmdargs.checknewer:
        # Force nocache mode
        cmdargs.nocache = True

    if cmdargs.subcmd == 'mbox' and cmdargs.refetch:
        return refetch(cmdargs.refetch)

    try:
        msgid, msgs = b4.retrieve_messages(cmdargs)
    except LookupError as ex:
        logger.critical('CRITICAL: %s', ex)
        sys.exit(1)

    if not msgs or not msgid:
        sys.exit(1)

    if len(msgs) and cmdargs.checknewer and b4.can_network:
        msgs = get_extra_series(msgs, direction=1, nocache=cmdargs.nocache)

    if cmdargs.subcmd in ('am', 'shazam'):
        make_am(msgs, cmdargs, msgid)
        return

    logger.info('%s messages in the thread', len(msgs))
    if cmdargs.subcmd == 'mbox' and cmdargs.minimize:
        msgs = minimize_thread(msgs)

    if cmdargs.outdir == '-':
        logger.info('---')
        save_msgs_as_mbox('-', msgs)
        return

    # Check if outdir is an existing maildir
    if b4.is_maildir(cmdargs.outdir):
        added = save_msgs_as_mbox(cmdargs.outdir, msgs, filterdupes=cmdargs.filterdupes)
        logger.info('Added %s messages to maildir %s', added, cmdargs.outdir)
        return

    config = b4.get_main_config()
    if cmdargs.maildir or config.get('save-maildirs', 'no') == 'yes':
        save_maildir = True
        dftext = 'maildir'
    else:
        save_maildir = False
        dftext = 'mbx'

    if cmdargs.wantname:
        savename = os.path.join(cmdargs.outdir, cmdargs.wantname)
    else:
        safe_msgid = re.sub(r'[^\w@.+%-]+', '_', msgid).strip('_')
        savename = os.path.join(cmdargs.outdir, f'{safe_msgid}.{dftext}')

    if save_maildir:
        if os.path.isdir(savename):
            shutil.rmtree(savename)
        mailbox.Maildir(savename, create=True)
        save_msgs_as_mbox(savename, msgs)
        logger.info('Saved maildir %s', savename)
        return

    save_msgs_as_mbox(savename, msgs)
    logger.info('Saved %s', savename)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2020 by the Linux Foundation
#
__author__ = 'Konstantin Ryabitsev <konstantin@linuxfoundation.org>'

import datetime
import json
import logging
import os
import pathlib
import shlex
import sqlite3

from email.message import EmailMessage
from typing import Any, Dict, List, Optional, Tuple

import requests

import b4

logger = logging.getLogger(__name__)

_STATUS_ORDER = {'pass': 0, 'warn': 1, 'fail': 2}

# In-process cache for sashiko API responses, keyed by message-id.
# This prevents redundant API calls when checking multiple patches
# from the same series within a single check run.
_sashiko_patchset_cache: Dict[str, Optional[Dict[str, Any]]] = {}


def clear_sashiko_cache() -> None:
    """Clear the sashiko patchset cache between check runs."""
    _sashiko_patchset_cache.clear()

SCHEMA_VERSION = 1

SCHEMA_SQL = '''
CREATE TABLE IF NOT EXISTS schema_version (
    version INTEGER PRIMARY KEY
);

CREATE TABLE IF NOT EXISTS check_results (
    msgid      TEXT NOT NULL,
    tool       TEXT NOT NULL,
    status     TEXT NOT NULL,
    summary    TEXT,
    url        TEXT,
    details    TEXT,
    checked_at TEXT NOT NULL DEFAULT (strftime('%Y-%m-%dT%H:%M:%S', 'now')),
    PRIMARY KEY (msgid, tool)
);
'''


# ---------------------------------------------------------------------------
# Cache database
# ---------------------------------------------------------------------------

def _get_db_path() -> str:
    """Return the path to the CI check cache database."""
    datadir = b4.get_data_dir()
    cidir = os.path.join(datadir, 'review')
    pathlib.Path(cidir).mkdir(parents=True, exist_ok=True)
    return os.path.join(cidir, 'ci.sqlite3')


def get_db() -> sqlite3.Connection:
    """Get a connection to the CI cache database, creating it if needed."""
    db_path = _get_db_path()
    is_new = not os.path.exists(db_path)
    conn = sqlite3.connect(db_path)
    conn.row_factory = sqlite3.Row
    if is_new:
        conn.executescript(SCHEMA_SQL)
        conn.execute(
            'INSERT OR REPLACE INTO schema_version (version) VALUES (?)',
            (SCHEMA_VERSION,))
        conn.commit()
    return conn


def cleanup_old(conn: sqlite3.Connection, max_days: int = 180) -> int:
    """Delete check results older than *max_days*. Returns count deleted."""
    cutoff = (datetime.datetime.now(datetime.timezone.utc)
              - datetime.timedelta(days=max_days)).isoformat()
    cursor = conn.execute(
        'DELETE FROM check_results WHERE checked_at < ?', (cutoff,))
    conn.commit()
    return cursor.rowcount


def get_cached_results(conn: sqlite3.Connection,
                       msgids: List[str]) -> Dict[str, List[Dict[str, str]]]:
    """Return cached check results keyed by msgid.

    Returns ``{msgid: [{tool, status, summary, url, details}, ...]}``.
    """
    if not msgids:
        return {}
    placeholders = ','.join('?' * len(msgids))
    cursor = conn.execute(
        'SELECT msgid, tool, status, summary, url, details'
        f' FROM check_results WHERE msgid IN ({placeholders})',
        msgids)
    results: Dict[str, List[Dict[str, str]]] = {}
    for row in cursor.fetchall():
        entry = {
            'tool': row['tool'],
            'status': row['status'],
            'summary': row['summary'] or '',
            'url': row['url'] or '',
            'details': row['details'] or '',
        }
        results.setdefault(row['msgid'], []).append(entry)
    return results


def store_results(conn: sqlite3.Connection, msgid: str,
                  results: List[Dict[str, str]]) -> None:
    """Store check results for a single message."""
    for entry in results:
        conn.execute(
            'INSERT OR REPLACE INTO check_results'
            ' (msgid, tool, status, summary, url, details)'
            ' VALUES (?, ?, ?, ?, ?, ?)',
            (msgid, entry['tool'], entry['status'],
             entry.get('summary', ''), entry.get('url', ''),
             entry.get('details', '')))
    conn.commit()


def delete_results(conn: sqlite3.Connection,
                   msgids: List[str]) -> None:
    """Delete all cached check results for the given message-ids."""
    if not msgids:
        return
    placeholders = ','.join('?' * len(msgids))
    conn.execute(
        f'DELETE FROM check_results WHERE msgid IN ({placeholders})',
        msgids)
    conn.commit()


# ---------------------------------------------------------------------------
# Config helpers
# ---------------------------------------------------------------------------

def load_check_cmds() -> Tuple[List[str], List[str]]:
    """Read check commands from git config.

    Returns ``(perpatch_cmds, series_cmds)`` as raw command strings.
    """
    config = b4.get_main_config()

    def _as_list(val: Any) -> List[str]:
        if isinstance(val, str):
            return [val]
        if isinstance(val, list):
            return list(val)
        return []

    perpatch = _as_list(config.get('review-perpatch-check-cmd'))
    if not perpatch:
        topdir = b4.git_get_toplevel()
        if topdir:
            checkpatch = os.path.join(topdir, 'scripts', 'checkpatch.pl')
            if os.access(checkpatch, os.X_OK):
                perpatch = ['_builtin_checkpatch']
    # Auto-wire patchwork CI when project is configured
    if '_builtin_patchwork' not in perpatch and config.get('pw-project') and config.get('pw-url'):
        perpatch.append('_builtin_patchwork')
    series = _as_list(config.get('review-series-check-cmd'))
    # Auto-wire sashiko AI review when URL is configured
    if config.get('sashiko-url'):
        if '_builtin_sashiko' not in perpatch:
            perpatch.append('_builtin_sashiko')
        if '_builtin_sashiko' not in series:
            series.append('_builtin_sashiko')
    return perpatch, series


def parse_cmd(cmdstr: str) -> List[str]:
    """Shell-split a command string into an argv list."""
    sp = shlex.shlex(cmdstr, posix=True)
    sp.whitespace_split = True
    return list(sp)


# ---------------------------------------------------------------------------
# Built-in handlers
# ---------------------------------------------------------------------------

def _run_builtin_checkpatch(msg: EmailMessage,
                            topdir: str) -> List[Dict[str, str]]:
    """Run scripts/checkpatch.pl on a single patch message."""
    checkpatch = os.path.join(topdir, 'scripts', 'checkpatch.pl')
    if not os.access(checkpatch, os.X_OK):
        return [{
            'tool': 'checkpatch',
            'status': 'fail',
            'summary': 'checkpatch.pl not found or not executable',
        }]

    cmdargs = [checkpatch, '-q', '--terse', '--no-summary', '--mailback']
    bdata = b4.LoreMessage.get_msg_as_bytes(msg)
    ecode, out, err = b4._run_command(cmdargs, stdin=bdata, rundir=topdir)

    out_str = out.strip().decode(errors='replace') if out and out.strip() else ''
    err_str = err.strip().decode(errors='replace') if err and err.strip() else ''

    findings: List[Dict[str, str]] = []
    worst = 'pass'
    for raw in (out_str.splitlines() + err_str.splitlines()):
        line = raw[2:] if raw.startswith('-:') else raw
        if not line:
            continue
        if 'ERROR:' in line:
            findings.append({'status': 'fail', 'description': line})
            worst = 'fail'
        elif 'WARNING:' in line or 'CHECK:' in line:
            findings.append({'status': 'warn', 'description': line})
            if worst != 'fail':
                worst = 'warn'
        else:
            # Continuation line — append to previous finding
            if findings:
                findings[-1]['description'] += ' ' + line
            else:
                findings.append({'status': 'pass', 'description': line})

    if not findings:
        if ecode:
            return [{
                'tool': 'checkpatch',
                'status': 'fail',
                'summary': f'exited with error code {ecode}',
            }]
        return [{
            'tool': 'checkpatch',
            'status': 'pass',
            'summary': 'passed all checks',
        }]

    errors = sum(1 for f in findings if f['status'] == 'fail')
    warnings = sum(1 for f in findings if f['status'] == 'warn')
    parts = []
    if errors:
        parts.append(f'{errors} error{"s" if errors != 1 else ""}')
    if warnings:
        parts.append(f'{warnings} warning{"s" if warnings != 1 else ""}')
    summary = ', '.join(parts) if parts else findings[0]['description']

    return [{
        'tool': 'checkpatch',
        'status': worst,
        'summary': summary,
        'details': json.dumps(findings),
    }]


def _run_builtin_patchwork(msg: EmailMessage, pwkey: str,
                           pwurl: str) -> List[Dict[str, str]]:
    """Query Patchwork REST API for checks on a single patch."""
    msgid = msg.get('message-id', '').strip('<> ')
    if not msgid:
        return []

    # Look up the PW patch ID for this message-id, then fetch its checks.
    try:
        pwdata = b4.LoreMessage.get_patchwork_data_by_msgid(msgid)
    except LookupError:
        logger.debug('Patchwork patch lookup failed for %s', msgid)
        return []
    patch_id = pwdata.get('id')
    if not patch_id:
        return []

    try:
        from b4.review import pw_fetch_checks
        checks = pw_fetch_checks(pwkey, pwurl, [int(patch_id)])
    except Exception as ex:
        logger.debug('Patchwork check query failed: %s', ex)
        return []

    if not checks:
        return []

    # Aggregate all PW checks into one result with worst-case status.
    worst = 'pass'
    individual: List[Dict[str, str]] = []
    counts: Dict[str, int] = {}
    for check in checks:
        state = check.get('state', 'pending')
        if state == 'success':
            status = 'pass'
        elif state in ('warning', 'pending'):
            status = 'warn'
        else:
            status = 'fail'
        if _STATUS_ORDER.get(status, 0) > _STATUS_ORDER.get(worst, 0):
            worst = status
        counts[status] = counts.get(status, 0) + 1
        individual.append({
            'context': check.get('context', 'unknown'),
            'status': status,
            'state': state,
            'description': check.get('description', ''),
            'url': check.get('url', ''),
        })

    summary_parts = []
    for s in ('pass', 'warn', 'fail'):
        if counts.get(s):
            summary_parts.append(f'{counts[s]} {s}')

    return [{
        'tool': 'patchwork',
        'status': worst,
        'summary': ', '.join(summary_parts),
        'details': json.dumps(individual),
    }]


def _fetch_sashiko_patchset(msgid: str, sashiko_url: str) -> Optional[Dict[str, Any]]:
    """Fetch patchset data from sashiko, with in-process caching.

    The cache ensures only one API call per series, even when checking
    multiple patches from the same patchset.
    """
    if msgid in _sashiko_patchset_cache:
        return _sashiko_patchset_cache[msgid]

    url = f'{sashiko_url.rstrip("/")}/api/patch'
    try:
        session = b4.get_requests_session()
        resp = session.get(url, params={'id': msgid}, timeout=30)
        if resp.status_code == 404:
            _sashiko_patchset_cache[msgid] = None
            return None
        resp.raise_for_status()
        data: Dict[str, Any] = resp.json()
    except requests.RequestException as ex:
        logger.debug('Sashiko API query failed for %s: %s', msgid, ex)
        _sashiko_patchset_cache[msgid] = None
        return None

    # Cache by all message-ids in this patchset so subsequent patches
    # in the same series hit the cache instead of the network.
    cover_msgid = data.get('message_id', '')
    if cover_msgid:
        _sashiko_patchset_cache[cover_msgid] = data
    for patch in data.get('patches', []):
        p_msgid = patch.get('message_id', '')
        if p_msgid:
            _sashiko_patchset_cache[p_msgid] = data
    _sashiko_patchset_cache[msgid] = data
    return data


def _parse_sashiko_findings(review: Dict[str, Any]) -> List[Dict[str, str]]:
    """Parse findings from a sashiko review's output JSON."""
    output_str = review.get('output', '') or ''
    if not output_str:
        return []
    try:
        output = json.loads(output_str)
    except (json.JSONDecodeError, TypeError):
        return []

    raw_findings = output.get('findings', [])
    findings: List[Dict[str, str]] = []
    for f in raw_findings:
        if not isinstance(f, dict):
            continue
        severity = (f.get('severity', '') or '').lower()
        problem = f.get('problem', f.get('title', ''))
        suggestion = f.get('suggestion', '')
        if severity in ('critical', 'high'):
            status = 'fail'
        elif severity == 'medium':
            status = 'warn'
        else:
            status = 'pass'
        desc = str(problem)
        if suggestion:
            desc += f' \u2014 {suggestion}'
        findings.append({
            'status': status,
            'context': f'sashiko/{severity}',
            'state': severity,
            'description': desc,
        })
    return findings


def _sashiko_findings_summary(findings: List[Dict[str, str]]) -> Tuple[str, str]:
    """Return ``(worst_status, summary_text)`` for a list of findings."""
    if not findings:
        return 'pass', 'No findings'

    worst = 'pass'
    counts: Dict[str, int] = {'critical': 0, 'high': 0, 'medium': 0, 'low': 0}
    for f in findings:
        state = f.get('state', '')
        if state in counts:
            counts[state] += 1
        if _STATUS_ORDER.get(f['status'], 0) > _STATUS_ORDER.get(worst, 0):
            worst = f['status']

    parts = []
    for sev in ('critical', 'high', 'medium', 'low'):
        if counts[sev]:
            parts.append(f'{counts[sev]} {sev}')
    return worst, ', '.join(parts)


def _run_builtin_sashiko(msg: EmailMessage,
                         sashiko_url: str) -> List[Dict[str, str]]:
    """Query sashiko AI review service for findings on a patch."""
    msgid = msg.get('message-id', '').strip('<> ')
    if not msgid:
        return []

    data = _fetch_sashiko_patchset(msgid, sashiko_url)
    if not data:
        return []

    ps_status = data.get('status', '')
    ps_id = data.get('id', '')
    reviews = data.get('reviews', [])
    patches = data.get('patches', [])
    base_url = sashiko_url.rstrip('/')
    patchset_url = f'{base_url}/patch/{ps_id}' if ps_id else ''

    # Build a map from patch message-id to sashiko patch id
    patch_id_by_msgid: Dict[str, int] = {}
    for p in patches:
        p_msgid = p.get('message_id', '')
        p_id = p.get('id')
        if p_msgid and p_id is not None:
            patch_id_by_msgid[p_msgid] = int(p_id)

    cover_msgid = data.get('message_id', '')
    is_cover = (msgid == cover_msgid)

    # Overall patchset status check (applies to cover letter row or
    # when the series is not yet reviewed).
    if ps_status in ('Pending', 'In Review', 'Applying'):
        return [{'tool': 'sashiko', 'status': 'warn',
                 'summary': f'Review {ps_status.lower()}',
                 'url': patchset_url}]
    if ps_status in ('Failed', 'Failed To Apply'):
        return [{'tool': 'sashiko', 'status': 'fail',
                 'summary': ps_status, 'url': patchset_url}]
    if ps_status == 'Incomplete':
        return [{'tool': 'sashiko', 'status': 'warn',
                 'summary': 'Series incomplete', 'url': patchset_url}]

    if is_cover:
        # Aggregate findings across all reviews for the cover letter
        all_findings: List[Dict[str, str]] = []
        for review in reviews:
            all_findings.extend(_parse_sashiko_findings(review))
        worst, summary = _sashiko_findings_summary(all_findings)
        result: Dict[str, str] = {
            'tool': 'sashiko', 'status': worst,
            'summary': summary, 'url': patchset_url,
        }
        if all_findings:
            result['details'] = json.dumps(all_findings)
        return [result]

    # Per-patch: find the matching review
    sashiko_patch_id = patch_id_by_msgid.get(msgid)
    if sashiko_patch_id is None:
        return []

    for review in reviews:
        if review.get('patch_id') == sashiko_patch_id:
            review_status = review.get('status', '')
            if review_status == 'Skipped':
                result_msg = review.get('result', '') or 'Skipped'
                return [{'tool': 'sashiko', 'status': 'pass',
                         'summary': result_msg, 'url': patchset_url}]
            if review_status in ('Pending', 'In Review'):
                return [{'tool': 'sashiko', 'status': 'warn',
                         'summary': 'Review in progress',
                         'url': patchset_url}]
            if review_status == 'Failed':
                result_msg = review.get('result', '') or 'Review failed'
                return [{'tool': 'sashiko', 'status': 'fail',
                         'summary': result_msg, 'url': patchset_url}]
            # Reviewed — parse findings
            findings = _parse_sashiko_findings(review)
            worst, summary = _sashiko_findings_summary(findings)
            result = {
                'tool': 'sashiko', 'status': worst,
                'summary': summary, 'url': patchset_url,
            }
            if findings:
                result['details'] = json.dumps(findings)
            return [result]

    # No review found for this patch
    return [{'tool': 'sashiko', 'status': 'pass',
             'summary': 'No review', 'url': patchset_url}]


# ---------------------------------------------------------------------------
# External command runner
# ---------------------------------------------------------------------------

def _run_external_cmd(cmdargs: List[str], msg: EmailMessage,
                      topdir: str,
                      extra_env: Optional[Dict[str, str]] = None,
                      ) -> List[Dict[str, str]]:
    """Run an external check command and parse its JSON output."""
    bdata = b4.LoreMessage.get_msg_as_bytes(msg)
    saved_env: Dict[str, Optional[str]] = {}
    if extra_env:
        for key, val in extra_env.items():
            saved_env[key] = os.environ.get(key)
            os.environ[key] = val
    try:
        ecode, out, err = b4._run_command(cmdargs, stdin=bdata, rundir=topdir)
    finally:
        for key, prev in saved_env.items():
            if prev is None:
                os.environ.pop(key, None)
            else:
                os.environ[key] = prev

    if not out or not out.strip():
        if ecode:
            mycmd = os.path.basename(cmdargs[0])
            err_msg = err.strip().decode(errors='replace') if err else ''
            return [{
                'tool': mycmd,
                'status': 'fail',
                'summary': f'exited with error code {ecode}',
                'details': err_msg,
            }]
        return []

    try:
        data = json.loads(out)
    except json.JSONDecodeError as ex:
        mycmd = os.path.basename(cmdargs[0])
        return [{
            'tool': mycmd,
            'status': 'fail',
            'summary': f'invalid JSON output: {ex}',
            'details': out.decode(errors='replace'),
        }]

    if not isinstance(data, list):
        data = [data]

    results: List[Dict[str, str]] = []
    for entry in data:
        if not isinstance(entry, dict):
            continue
        tool = entry.get('tool', os.path.basename(cmdargs[0]))
        status = entry.get('status', 'fail')
        if status not in ('pass', 'warn', 'fail'):
            status = 'fail'
        results.append({
            'tool': tool,
            'status': status,
            'summary': entry.get('summary', ''),
            'url': entry.get('url', ''),
            'details': entry.get('details', ''),
        })
    return results


# ---------------------------------------------------------------------------
# High-level runners
# ---------------------------------------------------------------------------

def _dispatch_cmd(cmdstr: str, msg: EmailMessage, topdir: str,
                  pwkey: str = '', pwurl: str = '',
                  extra_env: Optional[Dict[str, str]] = None,
                  ) -> List[Dict[str, str]]:
    """Run a single check command (built-in or external) against a message."""
    if cmdstr == '_builtin_checkpatch':
        return _run_builtin_checkpatch(msg, topdir)
    if cmdstr == '_builtin_patchwork':
        if pwkey and pwurl:
            return _run_builtin_patchwork(msg, pwkey, pwurl)
        logger.debug('_builtin_patchwork requested but pw-key/pw-url not configured')
        return []
    if cmdstr == '_builtin_sashiko':
        config = b4.get_main_config()
        sashiko_url = str(config.get('sashiko-url', ''))
        if sashiko_url:
            return _run_builtin_sashiko(msg, sashiko_url)
        logger.debug('_builtin_sashiko requested but sashiko-url not configured')
        return []

    cmdargs = parse_cmd(cmdstr)
    return _run_external_cmd(cmdargs, msg, topdir, extra_env=extra_env)


def run_perpatch_checks(
        patches: List[Tuple[str, EmailMessage]],
        cmds: List[str],
        topdir: str,
        pwkey: str = '',
        pwurl: str = '',
        extra_env: Optional[Dict[str, str]] = None,
) -> Dict[str, List[Dict[str, str]]]:
    """Run per-patch check commands on each patch.

    *patches* is a list of ``(msgid, EmailMessage)`` tuples.

    Returns ``{msgid: [result, ...]}``.
    """
    results: Dict[str, List[Dict[str, str]]] = {}
    for msgid, msg in patches:
        patch_results: List[Dict[str, str]] = []
        for cmdstr in cmds:
            try:
                patch_results.extend(_dispatch_cmd(cmdstr, msg, topdir,
                                                   pwkey, pwurl,
                                                   extra_env=extra_env))
            except Exception as ex:
                logger.debug('Check command %s failed: %s', cmdstr, ex)
                patch_results.append({
                    'tool': cmdstr.split()[0] if cmdstr else 'unknown',
                    'status': 'fail',
                    'summary': str(ex),
                })
        results[msgid] = patch_results
    return results


def run_series_checks(
        cover_msg: Tuple[str, EmailMessage],
        cmds: List[str],
        topdir: str,
        pwkey: str = '',
        pwurl: str = '',
        extra_env: Optional[Dict[str, str]] = None,
) -> List[Dict[str, str]]:
    """Run per-series check commands on the cover letter.

    *cover_msg* is a ``(msgid, EmailMessage)`` tuple.

    Returns ``[result, ...]``.
    """
    _msgid, msg = cover_msg
    results: List[Dict[str, str]] = []
    for cmdstr in cmds:
        try:
            results.extend(_dispatch_cmd(cmdstr, msg, topdir,
                                         pwkey, pwurl,
                                         extra_env=extra_env))
        except Exception as ex:
            logger.debug('Series check command %s failed: %s', cmdstr, ex)
            results.append({
                'tool': cmdstr.split()[0] if cmdstr else 'unknown',
                'status': 'fail',
                'summary': str(ex),
            })
    return results



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2024 by the Linux Foundation
#
__author__ = 'Konstantin Ryabitsev <konstantin@linuxfoundation.org>'

import email.utils

from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

import b4
import b4.review
import b4.review.tracking

from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Vertical
from textual.screen import ModalScreen
from textual.widgets import Label, ListItem, ListView, LoadingIndicator, RichLog, Static
from textual.worker import Worker, WorkerState

from rich.text import Text

from b4.review_tui._common import (
    resolve_styles, _quiet_worker, _fix_ansi_theme,
    _write_diff_line, display_width, pad_display,
)
from b4.review_tui._modals import FollowupReplyPreviewScreen


@dataclass
class ThreadNode:
    """A single message in the thread tree."""
    lmsg: b4.LoreMessage
    children: List['ThreadNode'] = field(default_factory=list)
    depth: int = 0
    tree_art: str = ''
    is_patch: bool = False
    attestation: List[Dict[str, Any]] = field(default_factory=list)
    att_passing: bool = True
    is_unseen: bool = False
    is_flagged: bool = False
    is_answered: bool = False


def _flatten_tree(
    roots: List[ThreadNode],
    prefix: str = '',
    is_root: bool = True,
) -> List[ThreadNode]:
    """DFS-flatten a list of roots into a list with tree_art set."""
    result: List[ThreadNode] = []
    for i, node in enumerate(roots):
        is_last = (i == len(roots) - 1)
        if is_root:
            node.tree_art = ''
        else:
            node.tree_art = prefix + ('\u2514\u2500>' if is_last else '\u251c\u2500>')
        result.append(node)
        if node.children:
            if is_root:
                child_prefix = ''
            else:
                child_prefix = prefix + ('  ' if is_last else '\u2502 ')
            result.extend(_flatten_tree(node.children, child_prefix, is_root=False))
    return result


def build_thread_tree(lmbx: b4.LoreMailbox) -> List[ThreadNode]:
    """Build a flat, DFS-ordered thread list from a LoreMailbox.

    Creates ThreadNode for every message, links children to parents
    via in_reply_to, sorts by date, flattens with tree art, and
    checks attestation status for each message.
    """
    config = b4.get_main_config()
    attpolicy = str(config.get('attestation-policy', 'softfail'))
    try:
        maxdays = int(str(config.get('attestation-staleness-days', '0')))
    except ValueError:
        maxdays = 0

    nodes: Dict[str, ThreadNode] = {}
    for msgid, lmsg in lmbx.msgid_map.items():
        att_list: List[Dict[str, Any]] = []
        att_passing = True
        if attpolicy != 'off':
            att_list, att_passing, _critical = lmsg.get_attestation_status(attpolicy, maxdays)
        nodes[msgid] = ThreadNode(
            lmsg=lmsg,
            is_patch=lmsg.has_diff,
            attestation=att_list,
            att_passing=att_passing,
        )

    roots: List[ThreadNode] = []
    for node in nodes.values():
        parent_id = node.lmsg.in_reply_to
        if parent_id and parent_id in nodes:
            nodes[parent_id].children.append(node)
        else:
            roots.append(node)

    def _sort_key(n: ThreadNode) -> Any:
        return n.lmsg.date

    def _sort_recursive(node_list: List[ThreadNode]) -> None:
        node_list.sort(key=_sort_key)
        for n in node_list:
            if n.children:
                _sort_recursive(n.children)

    _sort_recursive(roots)

    flat = _flatten_tree(roots)
    for i, node in enumerate(flat):
        node.depth = i  # store position for reference
    return flat


def _build_thread_label(node: ThreadNode,
                        ts: Optional[Dict[str, str]] = None) -> Text:
    """Build the Text label for a thread index row."""
    lmsg = node.lmsg
    if lmsg.date:
        date_str = lmsg.date.strftime('%d %b')
    else:
        date_str = '     '
    author = lmsg.fromname or lmsg.fromemail
    if display_width(author) > 20:
        while display_width(author) > 19:
            author = author[:-1]
        author += '\u2026'
    author = pad_display(author, 20)
    is_unseen = node.is_unseen
    unseen_style = f"bold {ts['warning']}" if ts else 'bold'
    flag_style = f"bold {ts['accent']}" if ts else 'bold'
    answered_style = ts['success'] if ts else ''
    if node.is_answered:
        row_style = f"dim {ts['success']}" if ts else 'dim'
    elif is_unseen:
        row_style = ''
    elif node.is_flagged:
        row_style = f"bold {ts['accent']}" if ts else 'bold'
    else:
        row_style = 'dim'
    text = Text(no_wrap=True, overflow='ellipsis')
    if node.is_answered:
        text.append('\u21a9 ', style=answered_style)
    elif is_unseen:
        text.append('N ', style=unseen_style)
    elif node.is_flagged:
        text.append('\u2605 ', style=flag_style)
    else:
        text.append('  ', style=row_style)
    text.append(f'{date_str} ', style=row_style)
    text.append(f'{author} ', style=row_style)
    if node.tree_art:
        text.append(node.tree_art, style='dim' if not node.is_flagged else row_style)
    text.append(lmsg.full_subject, style=row_style)
    return text


class ThreadIndexItem(ListItem):
    """A single row in the thread index showing date, author, tree art and subject."""

    def __init__(self, node: ThreadNode) -> None:
        super().__init__()
        self.node = node

    def compose(self) -> ComposeResult:
        ts = resolve_styles(self.app)
        yield Label(_build_thread_label(self.node, ts))


class MessageViewScreen(ModalScreen[None]):
    """Full message view with headers and body."""

    BINDINGS = [
        Binding('r', 'reply', 'reply'),
        Binding('F', 'toggle_flag', 'flag', key_display='F'),
        Binding('S', 'skip_quoted', 'skip quoted'),
        Binding('j', 'next_message', 'next msg'),
        Binding('k', 'prev_message', 'prev msg'),
        Binding('q', 'back', 'back'),
        Binding('escape', 'back', 'back', show=False),
        Binding('enter', 'scroll_down', 'scroll down', show=False),
        Binding('backspace', 'scroll_up', 'scroll up', show=False),
        Binding('space', 'page_down', 'page down', show=False),
        Binding('minus', 'page_up', 'page up', show=False),
        Binding('circumflex_accent', 'scroll_top', 'top', show=False),
        Binding('dollar_sign', 'scroll_bottom', 'bottom', show=False),
    ]

    DEFAULT_CSS = """
    MessageViewScreen {
        align: center middle;
    }
    #msg-dialog {
        width: 92%;
        height: 90%;
        border: solid $accent;
        background: $surface;
        padding: 1 2;
    }
    #msg-title {
        text-style: bold;
    }
    #msg-viewer {
        height: 1fr;
    }
    #msg-hint {
        height: 1;
        dock: bottom;
        color: $text-muted;
    }
    """

    def __init__(self, node: ThreadNode, lite_screen: 'LiteThreadScreen') -> None:
        super().__init__()
        self._node = node
        self._lite_screen = lite_screen

    def compose(self) -> ComposeResult:
        with Vertical(id='msg-dialog'):
            yield Static(f'Subject: {self._node.lmsg.full_subject}', id='msg-title', markup=False)
            yield RichLog(id='msg-viewer', highlight=False, wrap=True,
                          markup=False, auto_scroll=False)
            yield Static(
                'r reply  |  F flag  |  S skip quoted  |  j/k prev/next msg  |  q back',
                id='msg-hint',
            )

    def on_mount(self) -> None:
        self._render_message()
        self._update_title()
        self._lite_screen.mark_seen(self._node)

    def _update_title(self) -> None:
        """Update the subject title bar, reflecting flagged state."""
        subject = self._node.lmsg.full_subject
        title = self.query_one('#msg-title', Static)
        if self._node.is_flagged:
            ts = resolve_styles(self.app)
            text = Text()
            text.append(f'Subject: {subject} \u2605', style=f"bold {ts['accent']}")
            title.update(text)
        else:
            title.update(f'Subject: {subject}')

    def _render_message(self) -> None:
        """Render the current node's message into the viewer."""
        lmsg = self._node.lmsg
        msg = lmsg.msg
        viewer = self.query_one('#msg-viewer', RichLog)
        viewer.clear()
        ts = resolve_styles(self.app)

        # Content width: 92% of terminal minus border (2) and padding (4)
        content_width = max(40, int(self.app.size.width * 0.92) - 6)

        # Render headers — dim so the body draws the eye
        addr_hdrs = {'to', 'cc'}
        for hdr_name in ('Date', 'From', 'To', 'Cc'):
            val = msg.get(hdr_name)
            if not val:
                continue
            val = str(val)
            if hdr_name.lower() in addr_hdrs:
                pairs = email.utils.getaddresses([val])
                if pairs:
                    self._write_addr_header(viewer, hdr_name, pairs, content_width)
            else:
                hdr_text = Text()
                hdr_text.append(f'{hdr_name}: ', style='dim bold')
                hdr_text.append(val, style='dim')
                viewer.write(hdr_text)

        # Link header — more useful to maintainers than raw Message-ID
        config = b4.get_main_config()
        linkmask = str(config.get('linkmask', ''))
        if '%s' in linkmask and lmsg.msgid:
            linkurl = linkmask % lmsg.msgid
            hdr_text = Text()
            hdr_text.append('Link: ', style='dim bold')
            hdr_text.append(linkurl, style=f"dim link {linkurl}")
            viewer.write(hdr_text)

        # Attestation status
        node = self._node
        if node.attestation:
            att_text = Text()
            att_text.append('Attestation: ', style='dim bold')
            for i, att in enumerate(node.attestation):
                if i > 0:
                    att_text.append(', ', style='dim')
                status = att.get('status', 'unknown')
                identity = att.get('identity', 'unknown')
                if att.get('passing'):
                    att_text.append(f'\u2713 {identity}', style=ts['success'])
                    if 'mismatch' in att:
                        att_text.append(f' (From: {att["mismatch"]})', style=ts['warning'])
                else:
                    if status == 'badsig':
                        att_text.append(f'\u2717 BADSIG: {identity}', style=ts['error'])
                    elif status == 'nokey':
                        att_text.append(f'\u2717 No key: {identity}', style=ts['warning'])
                    else:
                        att_text.append(f'\u2717 {status}: {identity}', style=ts['error'])
            viewer.write(att_text)

        viewer.write('')

        # Render body
        body = lmsg.body or ''
        in_diff = False
        for line in body.splitlines():
            if line.startswith('diff --git '):
                in_diff = True
            if in_diff:
                _write_diff_line(viewer, line, ts=ts)
            elif line.startswith('>'):
                viewer.write(Text(line, style=f"dim {ts['accent']}"))
            elif line.startswith('---'):
                viewer.write(Text(line, style='dim'))
            else:
                viewer.write(Text(line))

    @staticmethod
    def _write_addr_header(
        viewer: RichLog, hdr_name: str,
        pairs: List[Any], width: int,
    ) -> None:
        """Write an address header, packing addresses to fill each line."""
        indent_len = len(hdr_name) + 2  # "Cc: "
        indent = ' ' * indent_len
        formatted = [b4.format_addrs([p], clean=True) for p in pairs]

        line = Text()
        line.append(f'{hdr_name}: ', style='dim bold')
        line_len = indent_len

        for i, addr in enumerate(formatted):
            sep = ', ' if i > 0 else ''
            needed = len(sep) + len(addr)
            if i > 0 and line_len + needed > width:
                # Finish current line and start a new indented one
                viewer.write(line)
                line = Text()
                line.append(indent, style='dim')
                line_len = indent_len
                sep = ''
                needed = len(addr)
            if sep:
                line.append(sep, style='dim')
            line.append(addr, style='dim')
            line_len += needed

        if line.plain.strip():
            viewer.write(line)

    def action_reply(self) -> None:
        self._lite_screen.compose_reply(self._node)

    def action_toggle_flag(self) -> None:
        self._lite_screen.toggle_flag(self._node)
        self._update_title()

    def _switch_to_node(self, offset: int) -> None:
        """Switch to the message at offset from current in the thread list."""
        nodes = self._lite_screen._thread_nodes
        try:
            idx = nodes.index(self._node)
        except ValueError:
            return
        new_idx = idx + offset
        if 0 <= new_idx < len(nodes):
            self._node = nodes[new_idx]
            self._update_title()
            self._render_message()
            self.query_one('#msg-viewer', RichLog).scroll_home()
            self._lite_screen.mark_seen(self._node)

    def action_next_message(self) -> None:
        self._switch_to_node(1)

    def action_prev_message(self) -> None:
        self._switch_to_node(-1)

    def action_skip_quoted(self) -> None:
        """Skip past the next block of quoted lines (mutt-style S)."""
        viewer = self.query_one('#msg-viewer', RichLog)
        lines = viewer.lines
        if not lines:
            return
        cur = int(viewer.scroll_y)
        total = len(lines)
        # Phase 1: find start of a quoted block at or after current pos
        i = cur
        while i < total and not lines[i].text.lstrip().startswith('>'):
            i += 1
        if i >= total:
            return  # no quoted block ahead
        # Phase 2: skip past the quoted block
        while i < total and lines[i].text.lstrip().startswith('>'):
            i += 1
        if i < total:
            # Show a couple of trailing quote lines for context
            target = max(0, i - 2)
            viewer.scroll_to(y=target)

    def action_back(self) -> None:
        self.dismiss(None)

    def action_scroll_down(self) -> None:
        self.query_one('#msg-viewer', RichLog).scroll_down()

    def action_scroll_up(self) -> None:
        self.query_one('#msg-viewer', RichLog).scroll_up()

    def action_page_down(self) -> None:
        self.query_one('#msg-viewer', RichLog).scroll_page_down()

    def action_page_up(self) -> None:
        self.query_one('#msg-viewer', RichLog).scroll_page_up()

    def action_scroll_top(self) -> None:
        self.query_one('#msg-viewer', RichLog).scroll_home()

    def action_scroll_bottom(self) -> None:
        self.query_one('#msg-viewer', RichLog).scroll_end()


class LiteThreadScreen(ModalScreen[None]):
    """Mutt-style lite thread viewer for browsing a mail thread."""

    DEFAULT_CSS = """
    LiteThreadScreen {
        align: center middle;
    }
    #lite-dialog {
        width: 92%;
        height: 90%;
        border: solid $accent;
        background: $surface;
        padding: 1 2;
    }
    #lite-list {
        height: 1fr;
    }
    #lite-loading {
        height: 1fr;
        content-align: center middle;
    }
    #lite-hint {
        height: 1;
        dock: bottom;
        color: $text-muted;
    }
    """

    BINDINGS = [
        Binding('j', 'cursor_down', 'down', show=False),
        Binding('k', 'cursor_up', 'up', show=False),
        Binding('q', 'back', 'back'),
        Binding('escape', 'back', 'back', show=False),
    ]

    def __init__(
        self,
        message_id: str,
        email_dryrun: bool = False,
        patatt_sign: bool = True,
        tracking_info: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__()
        self._message_id = message_id
        self._email_dryrun = email_dryrun
        self._patatt_sign = patatt_sign
        self._tracking_info = tracking_info
        self._thread_nodes: List[ThreadNode] = []

    def compose(self) -> ComposeResult:
        with Vertical(id='lite-dialog'):
            yield LoadingIndicator(id='lite-loading')
            yield Static(
                'Enter view  |  j/k navigate  |  q back',
                id='lite-hint',
            )

    def on_mount(self) -> None:
        _fix_ansi_theme(self.app)
        self.run_worker(self._fetch_thread, name='_fetch_thread', thread=True)

    def _refresh_msg_count(self, total_messages: int) -> None:
        """Opportunistically refresh message count from fetched messages."""
        if not self._tracking_info:
            return
        try:
            b4.review.tracking.refresh_message_count(
                self._tracking_info['identifier'],
                self._tracking_info['change_id'],
                self._tracking_info['revision'],
                total_messages,
            )
        except Exception:
            pass

    def _sync_unseen(self, unseen_count: int) -> None:
        """Sync seen_message_count in the tracking DB from unseen count."""
        if not self._tracking_info:
            return
        try:
            b4.review.tracking.sync_seen_from_unseen_count(
                self._tracking_info['identifier'],
                self._tracking_info['change_id'],
                self._tracking_info['revision'],
                unseen_count,
            )
        except Exception:
            pass

    def _fetch_thread(self) -> List[ThreadNode]:
        with _quiet_worker():
            msgs = b4.review._retrieve_messages(self._message_id)
            self._refresh_msg_count(len(msgs))
            lmbx = b4.LoreMailbox()
            for msg in msgs:
                lmbx.add_message(msg)
            return build_thread_tree(lmbx)

    async def on_worker_state_changed(self, event: Worker.StateChanged) -> None:
        if event.worker.name != '_fetch_thread':
            return
        if event.state == WorkerState.SUCCESS:
            if event.worker.result is None:
                return
            self._thread_nodes = event.worker.result
            await self._populate()
        elif event.state == WorkerState.ERROR:
            await self.query_one('#lite-loading', LoadingIndicator).remove()
            self.app.notify(str(event.worker.error), severity='error')

    async def _populate(self) -> None:
        await self.query_one('#lite-loading', LoadingIndicator).remove()
        if not self._thread_nodes:
            self.app.notify('No messages found', severity='warning')
            return
        try:
            from b4.review import messages
            conn = messages.get_db()
            msgids = [n.lmsg.msgid for n in self._thread_nodes if n.lmsg.msgid]
            flags_map = messages.get_flags_bulk(conn, msgids)
            conn.close()
            unseen_count = 0
            for node in self._thread_nodes:
                flags = flags_map.get(node.lmsg.msgid, '').split()
                node.is_unseen = 'Seen' not in flags
                node.is_flagged = 'Flagged' in flags
                node.is_answered = 'Answered' in flags
                if node.is_unseen:
                    unseen_count += 1
            # Sync the tracking DB so the Fups column matches the N markers
            self._sync_unseen(unseen_count)
        except Exception:
            pass
        self._detect_maintainer_replies()
        items = [ThreadIndexItem(node) for node in self._thread_nodes]
        lv = ListView(*items, id='lite-list')
        await self.mount(lv, before=self.query_one('#lite-hint', Static))
        lv.focus()

    @staticmethod
    def _msg_date(node: ThreadNode) -> Optional[str]:
        return node.lmsg.date.isoformat() if node.lmsg.date else None

    def mark_seen(self, node: ThreadNode) -> None:
        """Mark a single message as Seen in the messages DB."""
        if not node.is_unseen:
            return
        node.is_unseen = False
        msgid = node.lmsg.msgid
        if not msgid:
            return
        try:
            from b4.review import messages
            conn = messages.get_db()
            messages.set_flag(conn, msgid, 'Seen', self._msg_date(node))
            conn.close()
        except Exception:
            pass

    def _mark_answered(self, node: ThreadNode) -> None:
        """Mark a message as Answered in the messages DB."""
        node.is_answered = True
        msgid = node.lmsg.msgid
        if not msgid:
            return
        try:
            from b4.review import messages
            conn = messages.get_db()
            messages.set_flag(conn, msgid, 'Answered', self._msg_date(node))
            conn.close()
        except Exception:
            pass

    def _detect_maintainer_replies(self) -> None:
        """Detect messages sent by the maintainer and infer flags.

        When the maintainer replied via their own email client:
        - their own messages are marked Seen
        - the immediate parent is marked Answered
        - all ancestor messages are marked Seen
        """
        try:
            _, maintainer_email = b4.get_mailfrom()
        except Exception:
            return
        if not maintainer_email:
            return
        maintainer_email = maintainer_email.lower()
        node_map = {n.lmsg.msgid: n for n in self._thread_nodes if n.lmsg.msgid}
        seen_entries: List[Dict[str, Optional[str]]] = []
        answered_entries: List[Dict[str, Optional[str]]] = []

        for node in self._thread_nodes:
            if not node.lmsg.fromemail:
                continue
            if node.lmsg.fromemail.lower() != maintainer_email:
                continue
            # Maintainer's own message → Seen
            if node.is_unseen:
                node.is_unseen = False
                seen_entries.append({'msgid': node.lmsg.msgid,
                                     'msg_date': self._msg_date(node)})
            # Immediate parent → Answered
            parent_id = node.lmsg.in_reply_to
            if parent_id and parent_id in node_map:
                parent = node_map[parent_id]
                if not parent.is_answered:
                    parent.is_answered = True
                    answered_entries.append({'msgid': parent_id,
                                             'msg_date': self._msg_date(parent)})
            # All ancestors → Seen
            ancestor_id = node.lmsg.in_reply_to
            while ancestor_id and ancestor_id in node_map:
                ancestor = node_map[ancestor_id]
                if ancestor.is_unseen:
                    ancestor.is_unseen = False
                    seen_entries.append({'msgid': ancestor_id,
                                         'msg_date': self._msg_date(ancestor)})
                ancestor_id = ancestor.lmsg.in_reply_to

        if not seen_entries and not answered_entries:
            return
        try:
            from b4.review import messages
            conn = messages.get_db()
            if seen_entries:
                messages.set_flags_bulk(conn, seen_entries, 'Seen')
            if answered_entries:
                messages.set_flags_bulk(conn, answered_entries, 'Answered')
            conn.close()
        except Exception:
            pass

    def toggle_flag(self, node: ThreadNode) -> None:
        """Toggle the Flagged state on a message."""
        node.is_flagged = not node.is_flagged
        msgid = node.lmsg.msgid
        if not msgid:
            return
        try:
            from b4.review import messages
            conn = messages.get_db()
            if node.is_flagged:
                messages.set_flag(conn, msgid, 'Flagged', self._msg_date(node))
            else:
                messages.remove_flag(conn, msgid, 'Flagged')
            conn.close()
        except Exception:
            pass

    def action_cursor_down(self) -> None:
        try:
            self.query_one('#lite-list', ListView).action_cursor_down()
        except Exception:
            pass

    def action_cursor_up(self) -> None:
        try:
            self.query_one('#lite-list', ListView).action_cursor_up()
        except Exception:
            pass

    def on_list_view_selected(self, event: ListView.Selected) -> None:
        """Handle Enter on thread index — open message view."""
        event.stop()
        if isinstance(event.item, ThreadIndexItem):
            self.app.push_screen(
                MessageViewScreen(event.item.node, self),
                callback=lambda _: self._refresh_labels(),
            )

    def _refresh_labels(self) -> None:
        """Refresh thread index labels to reflect flag/seen changes."""
        try:
            lv = self.query_one('#lite-list', ListView)
        except Exception:
            return
        ts = resolve_styles(self.app)
        for item in lv.query(ThreadIndexItem):
            item.query_one(Label).update(_build_thread_label(item.node, ts))

    def compose_reply(self, node: ThreadNode, initial_text: Optional[str] = None) -> None:
        """Compose a reply to the given thread node using external editor."""
        lmsg = node.lmsg
        if initial_text is not None:
            editor_text = initial_text
        else:
            orig_date = lmsg.date.strftime('%Y-%m-%d %H:%M %z') if lmsg.date else ''
            orig_author = lmsg.fromname or lmsg.fromemail
            body = lmsg.body or ''
            quoted = '\n'.join(f'> {line}' for line in body.splitlines())
            editor_text = f'On {orig_date}, {orig_author} wrote:\n{quoted}\n\n'

        with self.app.suspend():
            result = b4.edit_in_editor(editor_text.encode(), filehint='reply.eml')
        reply_text = result.decode(errors='replace')
        if reply_text == editor_text:
            self.app.notify('No changes made')
            return

        # Build entry dict that FollowupReplyPreviewScreen expects
        entry: Dict[str, Any] = {
            'lmsg': lmsg,
            'fromname': lmsg.fromname,
            'fromemail': lmsg.fromemail,
            'date': lmsg.date,
            'body': lmsg.body or '',
        }

        def _on_preview(action: Optional[str]) -> None:
            if action == 'send':
                self._send_reply(node, reply_text)
            elif action == 'edit':
                self.compose_reply(node, reply_text)

        self.app.push_screen(FollowupReplyPreviewScreen(entry, reply_text), _on_preview)

    def _send_reply(self, node: ThreadNode, text: str) -> None:
        """Build and send a reply to the given thread node."""
        lmsg = node.lmsg
        msg = lmsg.make_reply(text)
        try:
            with self.app.suspend():
                smtp, fromaddr = b4.get_smtp(dryrun=self._email_dryrun)
                sent = b4.send_mail(smtp, [msg], fromaddr=fromaddr,
                                    patatt_sign=self._patatt_sign,
                                    dryrun=self._email_dryrun,
                                    output_dir=None, reflect=False)
            if sent is None:
                self.app.notify('Failed to send reply.', severity='error')
            elif self._email_dryrun:
                self.app.notify(f'Dry-run: reply to {lmsg.fromemail} logged, not sent')
                self._mark_answered(node)
            else:
                self.app.notify(f'Reply sent to {lmsg.fromemail}')
                self._mark_answered(node)
        except Exception as ex:
            self.app.notify(f'Send failed: {ex}', severity='error')

    def action_back(self) -> None:
        if self._thread_nodes:
            unseen = sum(1 for n in self._thread_nodes if n.is_unseen)
            self._sync_unseen(unseen)
        self.dismiss(None)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2024 by the Linux Foundation
#
__author__ = 'Konstantin Ryabitsev <konstantin@linuxfoundation.org>'

import json
import pathlib

from typing import Any, Dict, List, Optional, Set, Tuple

import b4
import b4.review
import b4.review.tracking

from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.widgets import Footer, Label, ListItem, ListView, LoadingIndicator, Static
from textual.worker import Worker, WorkerState

from rich.text import Text

from b4.review_tui._common import resolve_styles, ci_styles, logger, SeparatedFooter, _fix_ansi_theme, pad_display
from b4.review_tui._modals import (
    CIChecksScreen, SetStateScreen, ApplyStateModal,
    LimitScreen, HelpScreen, PW_HELP_LINES,
)


def _format_series_label(series: Dict[str, Any], tracked: bool,
                         ts: Optional[Dict[str, str]] = None,
                         show_delegate: bool = True) -> Text:
    """Build a Text label for a series row.

    *ts* is a resolved theme styles dict from :func:`resolve_styles`.
    When *show_delegate* is ``False`` the delegate column is omitted.
    """
    track_mark = 'T' if tracked else ' '
    ci_state = series.get('check') or 'pending'
    ci_map = ci_styles(ts) if ts else {
        'pending': 'dim', 'success': 'green', 'warning': 'red', 'fail': 'bold red',
    }
    ci_style = ci_map.get(ci_state, ci_map['pending'])
    date = (series.get('date') or '')[:10]
    state = f"{(series.get('state') or 'new'):<15s}"
    submitter = pad_display(series.get('submitter') or 'Unknown', 30)
    name = series.get('name') or '(no subject)'
    text = Text()
    text.append(track_mark)
    text.append('\u25cf', style=ci_style)
    if show_delegate:
        delegate = pad_display(series.get('delegate') or '', 15)
        text.append(f' {date}  {state} {submitter} {delegate} {name}')
    else:
        text.append(f' {date}  {state} {submitter} {name}')
    return text


class PwSeriesItem(ListItem):
    """A single Patchwork series entry in the listing."""

    ACTION_REQUIRED_STATES = ('new', 'under-review')

    def __init__(self, series: Dict[str, Any], tracked: bool = False,
                 show_delegate: bool = True) -> None:
        super().__init__()
        self.series = series
        self.tracked = tracked
        self.show_delegate = show_delegate
        state = series.get('state', 'new')
        if state not in self.ACTION_REQUIRED_STATES:
            self.add_class('--dimmed')
        if tracked:
            self.add_class('--tracked')

    def compose(self) -> ComposeResult:
        ts = resolve_styles(self.app)
        yield Label(_format_series_label(self.series, self.tracked, ts,
                                         show_delegate=self.show_delegate),
                    markup=False)


class PwApp(App[None]):
    """Textual app for browsing Patchwork series."""

    TITLE = 'b4 review pw'

    DEFAULT_CSS = """
    PwApp {
        layout: vertical;
    }
    #pw-title {
        dock: top;
        width: 100%;
        height: 1;
        background: $primary-darken-2;
        color: $text;
        text-style: bold;
        content-align: left middle;
    }
    #pw-header {
        width: 100%;
        height: 1;
        background: $surface;
        color: $text-muted;
    }
    #pw-list {
        height: 1fr;
    }
    PwSeriesItem.--dimmed Label {
        color: $text-disabled;
    }
    PwSeriesItem.--hidden Label {
        color: $text-disabled;
        text-style: dim italic;
    }
    PwSeriesItem.--tracked Label {
        color: $success-lighten-2;
        text-style: bold;
    }
    PwSeriesItem.--tracked.--dimmed Label {
        color: $success;
    }
    #pw-loading {
        height: 1fr;
        content-align: center middle;
    }
    PwApp:ansi #pw-title {
        background: ansi_bright_black;
        color: ansi_default;
        text-style: bold;
    }
    PwApp:ansi #pw-header {
        background: ansi_default;
        color: ansi_default;
        text-style: dim;
    }
    PwApp:ansi PwSeriesItem.--dimmed Label {
        color: ansi_default;
        text-style: dim;
    }
    PwApp:ansi PwSeriesItem.--hidden Label {
        color: ansi_default;
        text-style: dim italic;
    }
    PwApp:ansi PwSeriesItem.--tracked Label {
        color: ansi_default;
        text-style: bold;
    }
    PwApp:ansi PwSeriesItem.--tracked.--dimmed Label {
        color: ansi_default;
    }
    """

    BINDING_GROUPS = {
        'view': 'Series', 'ci_checks': 'Series', 'track_series': 'Series',
        'set_state': 'Series', 'hide_series': 'Series',
        'refresh': 'App', 'limit': 'App', 'toggle_show_hidden': 'App', 'quit': 'App', 'help': 'App',
    }

    BINDINGS = [
        # Hidden navigation bindings
        Binding('j', 'cursor_down', 'Down', show=False),
        Binding('k', 'cursor_up', 'Up', show=False),
        Binding('u', 'unhide_series', 'unhide', show=False),
        # Series-specific actions
        Binding('c', 'ci_checks', 'ci checks'),
        Binding('t', 'track_series', 'track'),
        Binding('s', 'set_state', 'set state'),
        Binding('h', 'hide_series', 'hide'),
        # App-global actions
        Binding('r', 'refresh', 'refresh'),
        Binding('l', 'limit', 'limit'),
        Binding('H', 'toggle_show_hidden', 'show hidden', key_display='H'),
        Binding('q', 'quit', 'quit'),
        Binding('question_mark', 'help', 'help', key_display='?'),
    ]

    def __init__(self, pwkey: str, pwurl: str, pwproj: str,
                 email_dryrun: bool = False, patatt_sign: bool = True) -> None:
        super().__init__()
        self._pwkey = pwkey
        self._pwurl = pwurl
        self._pwproj = pwproj
        self._email_dryrun = email_dryrun
        self._patatt_sign = patatt_sign
        self._states: List[Dict[str, Any]] = []
        self._hidden_ids: Set[int] = set()
        self._tracked_ids: Set[int] = set()
        self._show_hidden: bool = False
        self._limit_pattern: str = ''
        self._all_series: List[Dict[str, Any]] = []
        self._tracking_identifier: Optional[str] = None
        self._tracking_enabled: bool = False
        self._load_local_data()
        self._load_tracking_data()

    def _get_local_data_path(self) -> pathlib.Path:
        datadir = b4.get_data_dir('patchwork')
        return pathlib.Path(datadir) / f'{self._pwproj}.json'

    def _load_local_data(self) -> None:
        path = self._get_local_data_path()
        if not path.exists():
            return
        try:
            with open(path, 'r', encoding='utf-8') as fp:
                data = json.load(fp)
            for sid, sdata in data.get('series', {}).items():
                if sdata.get('hidden'):
                    self._hidden_ids.add(int(sid))
        except (json.JSONDecodeError, OSError):
            pass

    def _load_tracking_data(self) -> None:
        # Try to resolve tracking identifier from current repo or use pwproj
        topdir = b4.git_get_toplevel()
        if topdir:
            self._tracking_identifier = b4.review.tracking.get_repo_identifier(topdir)
        if not self._tracking_identifier:
            # Fall back to patchwork project name
            self._tracking_identifier = self._pwproj
        if self._tracking_identifier and b4.review.tracking.db_exists(self._tracking_identifier):
            self._tracking_enabled = True
            self._tracked_ids = b4.review.tracking.get_tracked_pw_series_ids(self._tracking_identifier)

    def _save_local_data(self) -> None:
        path = self._get_local_data_path()
        data: Dict[str, Any] = {'series': {}}
        for sid in self._hidden_ids:
            data['series'][str(sid)] = {'hidden': True}
        try:
            with open(path, 'w', encoding='utf-8') as fp:
                json.dump(data, fp, indent=2)
        except OSError:
            pass

    def compose(self) -> ComposeResult:
        yield Static(' Patchwork — loading\u2026', id='pw-title', markup=False)
        yield LoadingIndicator(id='pw-loading')
        yield SeparatedFooter()

    def on_mount(self) -> None:
        _fix_ansi_theme(self)
        self.run_worker(self._fetch_initial, name='_fetch_initial', thread=True)

    def _fetch_initial(self) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
        import b4.review
        from b4.review_tui._common import _quiet_worker

        with _quiet_worker():
            series = b4.review.pw_fetch_series(self._pwkey, self._pwurl, self._pwproj)
            states = b4.review.pw_fetch_states(self._pwkey, self._pwurl, self._pwproj)
            return series, states

    async def on_worker_state_changed(self, event: Worker.StateChanged) -> None:
        if event.worker.name == '_fetch_initial':
            if event.state == WorkerState.SUCCESS:
                if event.worker.result is None:
                    return
                series_list, states = event.worker.result
                self._states = states
                await self._populate(series_list)
            elif event.state == WorkerState.ERROR:
                for widget in self.query('#pw-loading'):
                    await widget.remove()
                self.query_one('#pw-title', Static).update(' Patchwork — error fetching series')
                self.notify(str(event.worker.error), severity='error')

    async def _populate(self, series_list: List[Dict[str, Any]]) -> None:
        loading = self.query('#pw-loading')
        for widget in loading:
            await widget.remove()
        self._all_series = series_list
        await self._refresh_list()

    async def _refresh_list(self) -> None:
        title = self.query_one('#pw-title', Static)
        widgets = list(self.query('#pw-header, #pw-list'))
        for widget in widgets:
            await widget.remove()
        visible = []
        hidden_count = 0
        for s in self._all_series:
            sid = s.get('id')
            is_hidden = sid in self._hidden_ids
            if is_hidden:
                hidden_count += 1
                if self._show_hidden:
                    visible.append((s, True))
            else:
                visible.append((s, False))
        if self._limit_pattern:
            visible = [
                (s, h) for s, h in visible
                if self._matches_limit(s, self._limit_pattern)
            ]
        limit_suffix = f', limit: {self._limit_pattern}' if self._limit_pattern else ''
        if hidden_count and not self._show_hidden:
            title.update(f' Patchwork — {len(visible)} series ({hidden_count} hidden{limit_suffix})')
        elif hidden_count and self._show_hidden:
            title.update(f' Patchwork — {len(visible)} series (showing {hidden_count} hidden{limit_suffix})')
        elif self._limit_pattern:
            title.update(f' Patchwork — {len(visible)} action-required series{limit_suffix}')
        else:
            title.update(f' Patchwork — {len(visible)} action-required series')
        if not visible:
            return
        # Hide the delegate column when all values are empty or identical
        delegates = {s.get('delegate', '') or '' for s, _h in visible}
        show_delegate = len(delegates) > 1
        if show_delegate:
            header_text = f'   {"Date":<12s}{"State":<15s} {"Submitter":<30s} {"Delegate":<15s} {"Series"}'
        else:
            header_text = f'   {"Date":<12s}{"State":<15s} {"Submitter":<30s} {"Series"}'
        header = Static(header_text, id='pw-header')
        items = []
        for s, is_hidden in visible:
            sid = s.get('id')
            is_tracked = sid in self._tracked_ids if sid else False
            item = PwSeriesItem(s, tracked=is_tracked,
                                show_delegate=show_delegate)
            if is_hidden:
                item.add_class('--hidden')
            items.append(item)
        lv = ListView(*items, id='pw-list')
        await self.mount(header, before=self.query_one(Footer))
        await self.mount(lv, before=self.query_one(Footer))
        lv.focus()

    async def action_refresh(self) -> None:
        """Re-fetch series from Patchwork and rebuild the list."""
        for widget in self.query('#pw-header, #pw-list'):
            await widget.remove()
        self.query_one('#pw-title', Static).update(' Patchwork — refreshing\u2026')
        await self.mount(LoadingIndicator(id='pw-loading'), before=self.query_one(Footer))
        self.run_worker(self._fetch_initial, name='_fetch_initial', thread=True)

    @staticmethod
    def _matches_limit(series: Dict[str, Any], pattern: str) -> bool:
        """Check whether *series* matches the limit *pattern*.

        Tokens are split on whitespace and combined with AND logic:
        - ``s:<substring>`` matches against the state field
        - ``d:<substring>`` matches against the delegate field
        - bare text matches against name or submitter
        """
        for token in pattern.lower().split():
            if token.startswith('s:'):
                needle = token[2:]
                if needle not in (series.get('state', '') or '').lower():
                    return False
            elif token.startswith('d:'):
                needle = token[2:]
                if needle not in (series.get('delegate', '') or '').lower():
                    return False
            else:
                if (token not in (series.get('name', '') or '').lower()
                        and token not in (series.get('submitter', '') or '').lower()):
                    return False
        return True

    def action_limit(self) -> None:
        delegates = {s.get('delegate', '') or '' for s in self._all_series}
        if len(delegates) > 1:
            hint = 'Prefixes: s:<state>  d:<delegate>'
        else:
            hint = 'Prefixes: s:<state>'
        self.push_screen(LimitScreen(self._limit_pattern, hint=hint),
                         callback=self._on_limit)

    async def _on_limit(self, result: Optional[str]) -> None:
        if result is None:
            return
        self._limit_pattern = result
        await self._refresh_list()

    def action_cursor_down(self) -> None:
        try:
            self.query_one('#pw-list', ListView).action_cursor_down()
        except Exception:
            pass

    def action_cursor_up(self) -> None:
        try:
            self.query_one('#pw-list', ListView).action_cursor_up()
        except Exception:
            pass

    def action_view(self) -> None:
        """View a Patchwork series thread in the lite thread viewer."""
        item = self._get_highlighted_item()
        if item is None:
            return
        msgid = item.series.get('msgid', '')
        if not msgid:
            self.notify('No message-id available for this series', severity='error')
            return
        from b4.review_tui._lite_app import LiteThreadScreen
        self.push_screen(LiteThreadScreen(msgid,
                                          email_dryrun=self._email_dryrun,
                                          patatt_sign=self._patatt_sign))

    def action_ci_checks(self) -> None:
        """View CI check details for the highlighted series."""
        item = self._get_highlighted_item()
        if item is None:
            return
        check = item.series.get('check') or 'pending'
        if check == 'pending':
            self.notify('No CI checks available for this series', severity='information')
            return
        self.push_screen(CIChecksScreen(self._pwkey, self._pwurl, item.series))

    def _get_highlighted_item(self) -> Optional['PwSeriesItem']:
        try:
            lv = self.query_one('#pw-list', ListView)
        except Exception:
            return None
        child = lv.highlighted_child
        if isinstance(child, PwSeriesItem):
            return child
        return None

    def action_set_state(self) -> None:
        if not self._states:
            self.notify('States not loaded yet', severity='warning')
            return
        item = self._get_highlighted_item()
        if item is None:
            return
        current_state = item.series.get('state', 'new')
        self.push_screen(
            SetStateScreen(self._states, current_state),
            callback=lambda result: self._on_set_state(result, item),
        )

    def _on_set_state(self, result: Optional[Tuple[str, bool]], item: 'PwSeriesItem') -> None:
        if result is None:
            return
        new_state, archived = result
        patch_ids = item.series.get('patch_ids', [])
        if not patch_ids:
            self.notify('No patch IDs for this series', severity='warning')
            return

        series_name = item.series.get('name', '(no subject)')
        self.push_screen(
            ApplyStateModal(
                self._pwkey, self._pwurl, patch_ids,
                new_state, archived, series_name
            ),
            callback=lambda res: self._on_apply_complete(res, item),
        )

    def _on_apply_complete(self, result: Tuple[int, int, str], item: 'PwSeriesItem') -> None:
        ok, fail, new_state = result
        if fail:
            self.notify(f'{ok} updated, {fail} failed', severity='warning')
        else:
            self.notify(f'{ok} patch(es) set to {new_state}', severity='information')
        item.series['state'] = new_state
        if new_state in PwSeriesItem.ACTION_REQUIRED_STATES:
            item.remove_class('--dimmed')
        else:
            item.add_class('--dimmed')
        ts = resolve_styles(self)
        item.query_one(Label).update(_format_series_label(
            item.series, item.tracked, ts,
            show_delegate=item.show_delegate))

    def action_track_series(self) -> None:
        if not self._tracking_enabled:
            self.notify('Repository not enrolled. Enroll with: b4 review enroll', severity='warning')
            return
        item = self._get_highlighted_item()
        if item is None:
            return
        sid = item.series.get('id')
        if sid is None:
            return
        if sid in self._tracked_ids:
            self.notify('Series already tracked', severity='information')
            return

        series_name = item.series.get('name', '(no subject)')
        msgid = item.series.get('msgid', '')
        if not msgid:
            self.notify('No message-id available for this series', severity='error')
            return

        pw_series_id: int = sid

        # Suspend UI while retrieving from lore (produces logging output)
        with self.suspend():
            logger.info('Retrieving series: %s', msgid)
            try:
                msgs = b4.review._retrieve_messages(msgid)
            except Exception as ex:
                logger.critical('Error retrieving series: %s', ex)
                return

        try:
            lser = b4.review._get_lore_series(msgs)
        except LookupError as ex:
            self.notify(str(ex), severity='error')
            return

        # Extract series metadata
        revision = lser.revision
        sender_name = lser.fromname
        sender_email = lser.fromemail
        num_patches = lser.expected
        subject = lser.subject

        # Get message-id from cover letter or first patch
        try:
            ref_msg = b4.review.get_reference_message(lser)
        except LookupError:
            self.notify('Could not find cover letter or first patch', severity='error')
            return

        if lser.change_id:
            change_id = lser.change_id
        else:
            date_prefix = ref_msg.date.strftime('%Y%m%d')
            slug = ref_msg.lsubject.get_slug(sep='-', with_counter=False)[:60]
            change_id = f'{date_prefix}-{slug}-{lser.fingerprint[:12]}'
            logger.info('No change-id found, generated: %s', change_id)

        message_id = ref_msg.msgid
        sent_at: Optional[str] = None
        if ref_msg.date:
            sent_at = ref_msg.date.isoformat()

        # Add to database
        assert self._tracking_identifier is not None
        conn = b4.review.tracking.get_db(self._tracking_identifier)
        b4.review.tracking.add_series_to_db(
            conn, change_id, revision, subject, sender_name, sender_email,
            sent_at, message_id, num_patches, pw_series_id)

        conn.close()

        # Update UI
        self._tracked_ids.add(pw_series_id)
        item.tracked = True
        item.add_class('--tracked')
        ts = resolve_styles(self)
        item.query_one(Label).update(_format_series_label(
            item.series, True, ts,
            show_delegate=item.show_delegate))
        self.notify(f'Started tracking: {series_name}', severity='information', timeout=3)

    async def action_hide_series(self) -> None:
        item = self._get_highlighted_item()
        if item is None:
            return
        sid = item.series.get('id')
        if sid is None:
            return
        if sid in self._hidden_ids:
            self.notify('Series already hidden', severity='information')
            return
        self._hidden_ids.add(sid)
        self._save_local_data()
        title = item.series.get('name', '(no subject)')
        self.notify(f'Hidden: {title}', timeout=3)
        await self._refresh_list()

    async def action_toggle_show_hidden(self) -> None:
        if not self._hidden_ids:
            self.notify('No hidden series', severity='information')
            return
        self._show_hidden = not self._show_hidden
        await self._refresh_list()

    async def action_unhide_series(self) -> None:
        item = self._get_highlighted_item()
        if item is None:
            return
        sid = item.series.get('id')
        if sid is None:
            return
        if sid not in self._hidden_ids:
            self.notify('Series not hidden', severity='information')
            return
        self._hidden_ids.discard(sid)
        self._save_local_data()
        title = item.series.get('name', '(no subject)')
        self.notify(f'Restored: {title}', timeout=3)
        await self._refresh_list()

    def check_action(self, action: str, parameters: Tuple[Any, ...]) -> Optional[bool]:
        if action == 'ci_checks':
            item = self._get_highlighted_item()
            if item is None:
                return False
            return (item.series.get('check') or 'pending') != 'pending'
        if action == 'track_series':
            if not self._tracking_enabled:
                return True  # Allow action so user sees the helpful error message
            item = self._get_highlighted_item()
            if item is None:
                return False
            sid = item.series.get('id')
            return sid not in self._tracked_ids
        if action == 'toggle_show_hidden':
            return bool(self._hidden_ids)
        if action == 'unhide_series':
            item = self._get_highlighted_item()
            if item is None:
                return False
            sid = item.series.get('id')
            return sid in self._hidden_ids
        if action == 'hide_series':
            item = self._get_highlighted_item()
            if item is None:
                return False
            sid = item.series.get('id')
            return sid not in self._hidden_ids
        return True

    def on_list_view_selected(self, event: ListView.Selected) -> None:
        self.action_view()

    def on_list_view_highlighted(self, event: ListView.Highlighted) -> None:
        self.refresh_bindings()

    def action_help(self) -> None:
        """Show keybinding help."""
        self.push_screen(HelpScreen(PW_HELP_LINES))

    async def action_quit(self) -> None:
        self.exit()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2020-2021 by the Linux Foundation
#
__author__ = 'Konstantin Ryabitsev <konstantin@linuxfoundation.org>'

import argparse
import os
import sys
import pathlib
import re

import b4

logger = b4.logger


def main(cmdargs: argparse.Namespace) -> None:
    if cmdargs.showkeys:
        _, msgs = b4.retrieve_messages(cmdargs)
        if not msgs:
            logger.info('No messages found in the thread.')
            sys.exit(0)
        logger.info('---')
        try:
            import patatt
        except ModuleNotFoundError:
            logger.info('--show-keys requires the patatt library')
            sys.exit(1)

        keydata = set()
        for msg in msgs:
            xdk = msg.get('x-developer-key')
            xds = msg.get('x-developer-signature')
            if not xdk or not xds:
                continue
            # grab the selector they used
            kdata = b4.LoreMessage.get_parts_from_header(xdk)
            sdata = b4.LoreMessage.get_parts_from_header(xds)
            algo = kdata.get('a')
            identity = kdata.get('i')
            selector = sdata.get('s', 'default')
            if algo == 'openpgp':
                keyinfo = kdata.get('fpr')
            elif algo == 'ed25519':
                keyinfo = kdata.get('pk')
            else:
                logger.debug('Unknown key type: %s', algo)
                continue
            keydata.add((identity, algo, selector, keyinfo))

        if not keydata:
            logger.info('No keys found in the thread.')
            sys.exit(0)
        krpath = os.path.join(b4.get_data_dir(), 'keyring')
        pgp = False
        ecc = False
        for identity, algo, selector, keyinfo in keydata:
            if not identity:
                logger.warning('No identity found for key %s %s %s', algo, selector, keyinfo)
                continue
            if not keyinfo:
                logger.warning('No keyinfo found for key %s %s %s', algo, selector, identity)
                continue
            keypath = patatt.make_pkey_path(algo, identity, selector)  # type: ignore[attr-defined]
            fullpath = os.path.join(krpath, keypath)
            if os.path.exists(fullpath):
                status = 'known'
            else:
                status = 'unknown'
                if algo == 'openpgp':
                    try:
                        uids = b4.get_gpg_uids(keyinfo)
                        if len(uids):
                            status = 'in default keyring'
                    except KeyError:
                        pass
            pathlib.Path(os.path.dirname(fullpath)).mkdir(parents=True, exist_ok=True)

            logger.info('%s: (%s)', identity, status)
            logger.info('    keytype: %s', algo)
            if algo == 'openpgp':
                pgp = True
                logger.info('      keyid: %s', keyinfo[-16:])
                logger.info('        fpr: %s', ':'.join(re.findall(r'.{4}', keyinfo)))
            else:
                ecc = True
                logger.info('     pubkey: %s', keyinfo)
            logger.info('     krpath: %s', keypath)
            logger.info('   fullpath: %s', fullpath)
        logger.info('---')
        if pgp:
            logger.info('For openpgp keys:')
            logger.info('    gpg --recv-key [keyid]')
            logger.info('    gpg -a --export [keyid] > [fullpath]')
        if ecc:
            logger.info('For ed25519 keys:')
            logger.info('    echo [pubkey] > [fullpath]')

        sys.exit(0)

    logger.info('This command is experimental. Try --show-keys [msgid].')
